// Telos AP phone interface (TWOx12, 2101, Nx12)
var apConnections=[];

/*
Telos Assistant Producer Protocol

available functions: (for help see this document: http://www.telos-systems.com/support/TLSAPX2a-API.pdf)
connection management
 void phoneUS.OpenConnection(cid, string host, string user, string pass);
 void phoneUS.CloseConnection(cid);
 void phoneUS.CloseAllConnections();
show management
 void phoneUS.EnumerateShows(cid);
 void phoneUS.EnumerateHostsForShow(cid, string showName);
 void phoneUS.AttachAsTalent(cid, sting hostName);
 void phoneUS.ChangeShow(cid, string showName, string showPass);
 void phoneUS.ConnectToShow(cid, string showName, string showPass, string showHost);
 void phoneUS.EnumerateDirectors(cid);
 void phoneUS.AttachToDirector(cid, string directorName);
 void phoneUS.AttachAsTalent(cid);
live operations
 bool phoneUS.PostBusyAllButton(cid);
 bool phoneUS.PostDelayDumpButton(cid, bool down);
 bool phoneUS.PostDialKey(cid, int key);
 bool phoneUS.PostDropButton(cid, int column);
 bool phoneUS.PostHoldButton(cid, int column);
 bool phoneUS.PostLineButton(cid, int line, int column);
 bool phoneUS.PostMuteButton(cid);
 bool phoneUS.PostNextButton(cid);
 bool phoneUS.PostRecordButton(cid);
 bool phoneUS.PostSpeakerButton(cid);
 bool phoneUS.PostTransferButton(cid);
 bool phoneUS.PostUpdateAll(cid);
 bool phoneUS.PostText(cid, int textType, int line, string text);
extensions
 bool phoneUS.EnblocDial(int channel, int line, int column, string number);
 bool phoneUS.MuteRinger(int channel, bool mute);
 bool phoneUS.AcousticAdapt(int channel, bool adapt);

available events:
 phConnected(cid);
 phDisconnected(cid);
 phEnumerateShowsCallback(cid, ...);
 phEnumerateShowsForHostCallback(cud, ...);
 phEnumerateHostsCallback(cid, ...);
 phEnumerateDirectorsCallback(cid, ...);
 phSendShowInfo(cid, ...);
 phSendMode(cid, ...);
 phSendShowChange(cid, ...);
 phSendRecord(cid, ...);
 phSendBusyAll(cid, ...);
 phSendTransfer(cid, ...);
 phSendMute(cid, ...);
 phSendRingMute(cid, ...);
 phSendDelayDump(cid, ...);
 phSendHandset(cid, ...);
 phSendSpeaker(cid, ...);
 phSendLineState(cid, ...);
 phSendSeizeLine(cid, ...);
 phSendRingLine(cid, ...);
 phSendDropLine(cid, ...);
 phSendDropLine(cid, ...);
 phSendSelectLine(cid, ...);
 phSendDialKey(cid, ...);
 phSendText(cid, ...);
 phSendError(cid, ...);
*/

// ------------------------- * Telos AP local protocol logic * -------------------------

function apAbortConnection(cid, reason)
{
	if(cid in apConnections)
	{
		log.error_f("ap%d: %s", cid, reason);
		phoneUS.CloseConnection(cid);
		delete apConnections[cid];
	}
}

// ------------------------- * Telos AP interface entry point * -------------------------

/*
** phConnected
**
** called from C code when connection to host is established
*/
function phConnected(cid)
{
	log.info_f("phConnected(%d)", cid);
	if(!(cid in apConnections)) return;

	var ap=apConnections[cid];
	if(ap.state!='conhost') return;

	if(ap.sh_name=="")
	{
		log.info_f("ap%d: Connected to '%s', discovering active show", cid, ap.host);
		ap.ashname=null;
		ap.state='shwlist';
		ap.sh_skipped=0;
		phoneUS.EnumerateShows(cid);
	}
	else
	{
		log.info_f("ap%d: Connected to '%s', connecting to show '%s'", cid, ap.host, ap.sh_name);
	/* Connecting to show, awaiting showInfo, or error */
		ap.state='conshow';
		phoneUS.ConnectToShow(cid, ap.sh_name, ap.sh_pass, ap.sh_host);
	}
}

/*
** phDisconnected
**
** called from C code when connection to host is broken (by timeout, by device or by our request)
*/
function phDisconnected(cid)
{
	log.info_f("phDisconnected(%d)", cid);
	if(!(cid in apConnections)) return;

	var ap=apConnections[cid];
	if(ap.state=='idle') return;
	ap.state='conhost';

// TODO: indicate connection loss
	log.info_f("ap%d: Connection lost", cid);
	phoneIndiStudio(cid, 0, 0, 0);
}

/*
** phSendError
**
** box's response to our command
** despite it's name its called in case of success too
*/
function phSendError(cid, errorNumber)
{
	if(errorNumber==99001) return; /* ping error response, ignore */
	if(!(cid in apConnections)) return;

	var ap=apConnections[cid];
	switch(ap.state)
	{
	case 'idle': return;
	case 'shwlist':
		if(errorNumber=='2001') apAbortConnection(cid, "Error getting show list");
		return;
	case 'conshow':
		if(errorNumber==5)
		{
			log.info_f("ap%d: Connected to show, attaching as talent", cid);
			ap.state='attach';
			phoneUS.AttachAsTalent(cid);
		}
		else if(errorNumber==5002)
			apAbortConnection(cid, "Error connecting to show: Invalid show password");
		else if(errorNumber==5001)
		{
			if(ap.sh_name!="")
			{
			/* our target show is not active, connect to active and change */
				log.info_f("ap%d: Show '%s' is not active, trying to switch shows", cid, ap.sh_name);
				ap.ashname=null;
				ap.state='shwlist';
				phoneUS.EnumerateShows(cid);
			}
			else
				apAbortConnection(cid, "Error connecting to active show");
		}
		return;
	case 'conashow':
		if(errorNumber==5)
		{
			log.info_f("ap%d: Connected to active show, changing show", cid);
			ap.state='chshow';
			phoneUS.ChangeShow(cid, ap.sh_name, ap.sh_pass);
		}
		else if(errorNumber==5001 || errorNumber==5002)
			apAbortConnection(cid, "Error connecting to active show");
		return;
	case 'chshow':
		if(errorNumber==4)
		{
			log.info_f("ap%d: Connected to show, attaching as talent", cid);
			ap.state='attach';
			phoneUS.AttachAsTalent(cid);
		}
		else if(errorNumber==4001 || errorNumber==4002)
			apAbortConnection(cid, "Error changing show");
		return;
	case 'attach':
		if(errorNumber==8)
		{
			log.info_f("ap%d: Attached as talent, working", cid);
			ap.state='working';
			phoneIndiStudio(cid, ap.studio.num_ln, ap.studio.num_hyb, ap.num_fixed);
			phoneUS.PostUpdateAll(cid);
		}
		else if(errorNumber==8001 || errorNumber==8002)
			apAbortConnection(cid, "Error attaching as talent");
		return;
	}
}

function phEnumerateShowsCallback(cid, showName, rqPass, isActive, hostName, isLast)
{
	if(!(cid in apConnections)) return;

	var ap=apConnections[cid];
	if(ap.state!='shwlist') return;

	if(ap.sh_name=="")
	{
	/* we don't care which show we will get */
		if(isActive && (!rqPass || ap.sh_pass!="") && (ap.sh_host=="" || hostName==ap.sh_host))
		{
			if(ap.ashname==null || ap.sh_skipped<=ap.sh_skip)
			{
				ap.ashname=showName;
				ap.ashhost=hostName;
			}
			ap.sh_skipped++;
		}

		if(isLast)
		{
			if(ap.ashname==null)
			{
			/* active show not found, just close connection */
				apAbortConnection(cid, "Unable to find active show");
				return;
			}
			log.info_f("ap%d: Connecting to active show '%s@%s'", cid, ap.ashname, ap.ashhost);
			ap.state='conshow';
			phoneUS.ConnectToShow(cid, ap.ashname, ap.sh_pass, ap.ashhost);
		}
	}
	else
	{
		if(ap.ashname==null && isActive && hostName==ap.sh_host)
			ap.ashname=showName;

		if(isLast)
		{
			if(ap.ashname==null)
			{
			/* active show not found, just close connection */
				apAbortConnection(cid, "Unable to find active show @ host "+ap.sh_host);
				return;
			}
			log.info_f("ap%d: Connecting to active show '%s'", cid, ap.ashname);
			ap.state='conashow';
			phoneUS.ConnectToShow(cid, ap.ashname, ap.sh_pass, ap.sh_host);
		}
	}
}

/*
** phSendShowInfo
**
** phone box is sending us information about the show we are connected to
*/
function phSendShowInfo(cid, maxLines, maxHybrids)
{
	if(!(cid in apConnections)) return;

	var ap=apConnections[cid];
	if(ap.state=='idle') return;

	log.info_f("ap%d: %d hybrids, %d lines", cid, maxHybrids, maxLines);
	ap.studio=
	{
		num_ln:  maxLines,   // number of lines
		num_hyb: maxHybrids, // number of hybrids
		ln_map:  []          // line state map
	};
	var n;
	for(n=0; n<maxLines; n++)
		ap.studio.ln_map[n]={state: 'NONE', hybrid: 0, remote: ""};
}

function phSendShowChange(cid, newShowName, oldShowName)
{
	if(!(cid in apConnections)) return;

	var ap=apConnections[cid];
	if(ap.state!='working') return;

	log.info_f("ap%d: Show changed from '%s' to '%s'", cid, oldShowName, newShowName);
	if(newShowName=="<show_disconnected>")
	{
	/* we are disconnected from show, initiate reconnect */
	// TODO: notify client
		if(ap.sh_name=="")
		{
			ap.ashname=null;
			ap.state='shwlist';
			ap.sh_skipped=0;
			phoneUS.EnumerateShows(cid);
		}
		else
		{
		/* Connecting to show, awaiting showInfo, or error */
			ap.state='conshow';
			phoneUS.ConnectToShow(cid, ap.sh_name, ap.sh_pass, ap.sh_host);
		}
	}
	else if(ap.sh_name!="")
		ap.sh_name=newShowName; /* update show name, use it in case of a connection loss */
}

function phSendLineState(cid, line, hybrid, lineState, elapsedTimeSec)
{
	if(!(cid in apConnections)) return;

	var ap=apConnections[cid];
	if(ap.state!='working') return;
	if(!(line in ap.studio.ln_map)) return;
	var ln=ap.studio.ln_map[line];

	if(lineState==0x1020) return; // Ignore shared_unavailable state
/* convert state and hybrid info */
	switch(lineState)
	{
	case 0x0000: // INVALID
	default:
		lineState='NONE';
		hybrid=0;
		break;
	case 0x0100: // RESERVED
	case 0x0200: // IDLE
		lineState='IDLE';
		hybrid=0;
		break;
	case 0x0301: // RINGING
		lineState='RINGING_IN';
		hybrid=0;
		break;
	case 0x0402: // HOLD
	case 0x0702: // TALENT_HOLD
		lineState='ON_HOLD';
		hybrid=0;
		break;
	case 0x0504: // READY_HOLD
	case 0x0604: // READY_NEXT
		lineState='ON_HOLD_READY';
		hybrid=0;
		break;
	case 0x0808: // ON_AIR
		lineState='ON_AIR';
		hybrid++;
		break;
	case 0x0908: // ON_AIR_LOCKED
		lineState='ON_AIR_LOCKED';
		hybrid++;
		break;
	case 0x0A10: // HANDSET
	case 0x0B10: // HANDSET_LOCKED
	case 0x0C10: // SPEAKER
	case 0x0D10: // SPEAKER_LOCKED
		lineState='ON_HANDSET';
		hybrid=0;
		break;
	case 0x0E20: // USED_ELSWHERE
		lineState='USED_EW';
		hybrid=0;
		break;
	case 0x0F20: // BUSY_ALL
		lineState='BUSY';
		hybrid=0;
		break;
	}
	if(hybrid>ap.studio.num_hyb)
	{
		lineState='USED_EW';
		hybrid=0;
	}

	if(ln.hybrid!=hybrid || ln.state!=lineState)
	{
		ln.hybrid=hybrid;
		ln.state=lineState;
		apIndiLineState(cid, line);
	}
}

/*
** phSendRingLine
**
** event: incoming call
*/
function phSendRingLine(cid, callerId, line)
{
	if(!(cid in apConnections)) return;

	var ap=apConnections[cid];
	if(ap.state!='working') return;
	if(!(line in ap.studio.ln_map)) return;
	var ln=ap.studio.ln_map[line];

	ln.remote=callerId;
	apIndiLineState(cid, line);
}

/*
** phSendDropLine
**
** event: call dropped
*/
function phSendDropLine(cid, callerId, line)
{
	if(!(cid in apConnections)) return;

	var ap=apConnections[cid];
	if(ap.state!='working') return;
	if(!(line in ap.studio.ln_map)) return;
	var ln=ap.studio.ln_map[line];

	ln.remote=""; /* simply clear remote number */
}

/*
** phSendSelectLine
**
** event: call answered
*/
function phSendSelectLine(cid, callerId, line, hybrid)
{
	if(!(cid in apConnections)) return;

	var ap=apConnections[cid];
	if(ap.state!='working') return;
	if(!(line in ap.studio.ln_map)) return;
	var ln=ap.studio.ln_map[line];

/* XXX is this needed? */
	ln.hybrid=hybrid;
	ln.remote=callerId;
}

/*
** phSendSeizeLine
**
** event: line is seized and ready for dialing
*/
function phSendSeizeLine(cid, line, hybrid)
{
	if(!(cid in apConnections)) return;

	var ap=apConnections[cid];
	if(ap.state!='working') return;
	if(!(line in ap.studio.ln_map)) return;
	var ln=ap.studio.ln_map[line];

	if('call_number' in ln)
	{
		var n;
	/* send number keys one by one */
		for(n=0; n<ln.call_number.length; n++)
		{
			host.sleep(100);
			phoneUS.PostDialKey(cid, ln.call_number.charCodeAt(n));
		}
		delete ln.call_number;
	}
}

function phSendDialKey(cid, sym)
{
	if(!(cid in apConnections)) return;
}

function phSendMode(cid, talent) {}
function phSendRecord(cid, on) {}
function phSendBusyAll(cid, on) {}
function phSendTransfer(cid, on) {}
function phSendMute(cid, on) {}
function phSendRingMute(cid, on) {}
function phSendDelayDump(cid, on) {}
function phSendHandset(cid, on) {}
function phSendSpeaker(cid, on) {}
function phSendAnswerLine(cid, callerId, line, hybrid) {}
function phSendText(cid, textType, text, line) {}

function phEnumerateShowsForHostCallback(cid, showName, isLast) {}
function phEnumerateHostsCallback(cid, hostName, isLast) {}
function phEnumerateDirectorsCallback(cid, directorName, isTalent, isFree, isLast) {}

// ------------------------- * LineState * -------------------------

function apIndiLineState(cid, line)
{
	var ap=apConnections[cid];
	var ln=ap.studio.ln_map[line];
	var hybrid, fix_state;

/* fisrt indicate slectable line state */
	phoneIndiLineState(cid, line, ln.state, ln.hybrid, ln.remote);
/* next check if the line belongs to fixed hybrid */
	for(hybrid in ap.fix_map)
		if(ap.fix_map[hybrid]==line)
	{
		if(ln.hybrid==0 || ln.hybrid==hybrid)
			fix_state=ln.state;  /* hybrid neutral states */
		else
			fix_state='USED_EW'; /* used elsewhere */
		phoneIndiFixedState(cid, parseInt(hybrid), fix_state, ln.remote);
	}
}

// ------------------------- * Telos AP protocol commands * -------------------------

/*
** apConnect
**
** Connection request from scripts
**
** cid   - connection id
** host  - Telos Phone Server host (ip address)
** login/pass - Login
** shname/shpass/shhost
**   here is the logic:
**  if(shname=="")
**    will search for any active shows on a device conforming to:
**     if shpass=="" will look for pasword-less shows
**     if shhost=="" will look for any host, if not will look for that specific host
**     if shskip>0 will skip that number of valid shows
**  if(shname!="")
**    will try to activate show / host pair
*/
function apConnect(cid, host, login, pass, shname, shpass, shhost, shskip)
{
	if(cid in apConnections)
	{
		if(apConnections[cid].state!='idle') phoneUS.CloseConnection(cid);
		delete apConnections[cid];
	}
	if(host=="" || !phoneUS.OpenConnection(cid, host, login, pass))
		return;
	apConnections[cid]=
	{
		state: 'conhost', // {idle, conhost, conshow, attach, working}
// + some special states in case we need to connect to inactive show:
/// shwlist  - getting list of shows to find out active
/// conashow - connecting to active show
/// chshow   - changing show to the requested one
// additional variable:
/// ashname  - active show name
		host: host,
		login:login,
		pass: pass,
		sh_name: shname,
		sh_pass: shpass,
		sh_host: shhost,
		sh_skip: shskip,
		studio: null,
		fix_map: [], // maps fixed lines to hybrids
		num_fixed: 0 // number of virtual fixed hybrids (including gaps, i.e. max number)
	};
}

function apCmndBlockAllToggle(cid)
{
	if(!(cid in apConnections)) return;
	var ap=apConnections[cid];

	if(ap.state!='working') return;
	phoneUS.PostBusyAllButton(cid);
}

function apCmndTransfer(cid)
{
	if(!(cid in apConnections)) return;
	var ap=apConnections[cid];

	if(ap.state!='working') return;
	phoneUS.PostTransferButton(cid);
}

function apCmndNext(cid)
{
	if(!(cid in apConnections)) return;
	var ap=apConnections[cid];

	if(ap.state!='working') return;
	phoneUS.PostNextButton(cid);
}

function apCmndMute(cid, mute)
{
	if(!(cid in apConnections)) return;
	var ap=apConnections[cid];

	if(ap.state!='working') return;
	phoneUS.MuteRinger(cid, mute);
}

function apCmndHybridDrop(cid, hybrid)
{
	if(!(cid in apConnections)) return;
	var ap=apConnections[cid];

	if(ap.state!='working') return;
	if(hybrid<=0 || hybrid>ap.studio.num_hyb) return;

	for(var line in ap.studio.ln_map)
	{
		var ln=ap.studio.ln_map[line];
		if(ln.state=='SEIZED') /* drop our virtual SEIZED state */
		{
			ln.state='IDLE';
			ln.hybrid=0;
			apIndiLineState(cid, parseInt(line));
		}
	}
	phoneUS.PostDropButton(cid, hybrid-1);
}

function apCmndHybridHold(cid, hybrid)
{
	if(!(cid in apConnections)) return;
	var ap=apConnections[cid];

	if(ap.state!='working') return;
	if(hybrid<=0 || hybrid>ap.studio.num_hyb) return;

	for(var line in ap.studio.ln_map)
	{
		var ln=ap.studio.ln_map[line];
		if(ln.state=='SEIZED') /* drop our virtual SEIZED state */
		{
			ln.state='IDLE';
			ln.hybrid=0;
			apIndiLineState(cid, parseInt(line));
		}
	}
	phoneUS.PostHoldButton(cid, hybrid-1);
}

function apCmndLineSeize(cid, line, hybrid)
{
	if(!(cid in apConnections)) return;
	var ap=apConnections[cid];

	if(ap.state!='working') return;
	if(hybrid<0 || hybrid>ap.studio.num_hyb) return;
	if(!(line in ap.studio.ln_map)) return;
	var ln=ap.studio.ln_map[line];

/* AP doesn't have real SEIZED state, so we need to emulate that */
	if(ln.state=='IDLE')
	{
	/* need to drop other SEIZED lines first */
		var n;
		for(n=0; n<ap.studio.num_ln; n++)
			if(ap.studio.ln_map[n].state=='SEIZED')
		{
			ap.studio.ln_map[n].state='IDLE';
			ap.studio.ln_map[n].hybrid=0;
			apIndiLineState(cid, n);
		}
		ln.state='SEIZED';
		ln.hybrid=hybrid;
		apIndiLineState(cid, line);
	}
	else if(ln.state=='SEIZED' && ln.hybrid!=hybrid)
	{
	/* switching seized hybrid */
		ln.hybrid=hybrid;
		apIndiLineState(cid, line);
	}
}

function apCmndLineDrop(cid, line)
{
	if(!(cid in apConnections)) return;
	var ap=apConnections[cid];

	if(ap.state!='working') return;
	if(!(line in ap.studio.ln_map)) return;
	var ln=ap.studio.ln_map[line];

	if(ln.state=='SEIZED') /* drop our virtual SEIZED state */
	{
		ln.state='IDLE';
		ln.hybrid=0;
		apIndiLineState(cid, line);
	}
}

function apCmndLineCall(cid, line, number, hybrid)
{
	if(!(cid in apConnections)) return;
	var ap=apConnections[cid];

	if(ap.state!='working') return;
	if(!(line in ap.studio.ln_map)) return;
	var ln=ap.studio.ln_map[line];
	if(hybrid==0 && ln.state=='SEIZED') hybrid=ln.hybrid;
	if(hybrid<=0 || hybrid>ap.studio.num_hyb) return;

	if(ln.state=='SEIZED' || ln.state=='IDLE')
	{
	/* try to use Enbloc dialing extension first */
		if(!phoneUS.EnblocDial(cid, line, hybrid-1, number))
		{
		/* AP doesn't support real enbloc dialing, so we need to emulate that */
			ln.call_number=number;
		/* pick line and wait for seized event before dialing */
			phoneUS.PostLineButton(cid, line, hybrid-1);
		}
	}
}

function apCmndLineTake(cid, line, hybrid)
{
	if(!(cid in apConnections)) return;
	var ap=apConnections[cid];

	if(ap.state!='working') return;
	if(hybrid<=0 || hybrid>ap.studio.num_hyb) return;
	if(!(line in ap.studio.ln_map)) return;
	var ln=ap.studio.ln_map[line];

	if(ln.state!='ON_AIR' || ln.hybrid!=hybrid) // check if line is not already taken on that fader
		phoneUS.PostLineButton(cid, line, hybrid-1);
}

function apCmndLineLock(cid, line)
{
	if(!(cid in apConnections)) return;
	var ap=apConnections[cid];

	if(ap.state!='working') return;
	if(!(line in ap.studio.ln_map)) return;
	var ln=ap.studio.ln_map[line];

	if(ln.state=='ON_AIR' && ln.hybrid>0) // check if line is on air then send line button with same hybrid
		phoneUS.PostLineButton(cid, line, ln.hybrid-1);
}

function apCmndLineUnlock(cid, line)
{
	if(!(cid in apConnections)) return;
	var ap=apConnections[cid];

	if(ap.state!='working') return;
	if(!(line in ap.studio.ln_map)) return;
	var ln=ap.studio.ln_map[line];

	if(ln.state=='ON_AIR_LOCKED' && ln.hybrid>0) // check if line is on air locked then send line button with same hybrid
		phoneUS.PostLineButton(cid, line, ln.hybrid-1);
}

/*
** apCmndFixedRegister
**
** bounds line to hybrid to form Emulated fixed line in AP connection
** hybrid is primary key (numbered from 1), line is numbered from 0
*/
function apCmndFixedRegister(cid, hybrid, line)
{
	if(!(cid in apConnections) || hybrid<1) return;
	var ap=apConnections[cid], p;

	if(line===null)
	{
		if(hybrid in ap.fix_map)
			delete ap.fix_map[hybrid];
	}
	else
		ap.fix_map[hybrid]=line;
/* notify on number fixed hybrids change */
	ap.num_fixed=0;
	for(p in ap.fix_map)
		if(p>ap.num_fixed)
			ap.num_fixed=parseInt(p);
	phoneIndiStudio(cid, ap.studio.num_ln, ap.studio.num_hyb, ap.num_fixed);
	for(p=0; p<ap.studio.num_ln; p++)
		apIndiLineState(cid, p);
}

function apCmndFixedSeize(cid, hybrid)
{
	if(!(cid in apConnections)) return;
	var ap=apConnections[cid];

	if(hybrid in ap.fix_map) apCmndLineSeize(cid, ap.fix_map[hybrid], hybrid);
}

function apCmndFixedDrop(cid, hybrid)
{
	if(!(cid in apConnections)) return;
	var ap=apConnections[cid];

	if(hybrid in ap.fix_map) apCmndHybridDrop(cid, hybrid);
}

function apCmndFixedHold(cid, hybrid)
{
	if(!(cid in apConnections)) return;
	var ap=apConnections[cid];

	if(hybrid in ap.fix_map) apCmndHybridHold(cid, hybrid);
}

function apCmndFixedCall(cid, hybrid, number)
{
	if(!(cid in apConnections)) return;
	var ap=apConnections[cid];

	if(hybrid in ap.fix_map) apCmndLineCall(cid, ap.fix_map[hybrid], number, hybrid);
}

function apCmndFixedTake(cid, hybrid)
{
	if(!(cid in apConnections)) return;
	var ap=apConnections[cid];

	if(hybrid in ap.fix_map) apCmndLineTake(cid, ap.fix_map[hybrid], hybrid);
}

function apCmndFixedLock(cid, hybrid)
{
	if(!(cid in apConnections)) return;
	var ap=apConnections[cid];

	if(hybrid in ap.fix_map) apCmndLineLock(cid, ap.fix_map[hybrid]);
}

function apCmndFixedUnlock(cid, hybrid)
{
	if(!(cid in apConnections)) return;
	var ap=apConnections[cid];

	if(hybrid in ap.fix_map) apCmndLineUnlock(cid, ap.fix_map[hybrid]);
}

function apCmndFixedGetState(cid, hybrid)
{
	var res={state:'NONE',remote:''};

	if(!(cid in apConnections)) return res;
	var ap=apConnections[cid];

	if(ap.state!='working') return res;
	if(!(hybrid in ap.fix_map)) return res;
	if(!(ap.fix_map[hybrid] in ap.studio.ln_map)) return res;
	var ln=ap.studio.ln_map[ap.fix_map[hybrid]];

	if(ln.hybrid==0 || ln.hybrid==hybrid)
		res.state=ln.state;  /* hybrid neutral states */
	else
		res.state='USED_EW'; /* used elsewhere */
	res.remote=ln.remote;
	return res;
}
