// available functions: (for help see this document: http://www.telos-systems.com/support/TLSAPX2a-API.pdf)

// connection management
// void phoneUS.OpenConnection(int channel, string host, string user, string pass);
// void phoneUS.CloseConnection(int channel);
// void phoneUS.CloseAllConnections();
// show management
// void phoneUS.EnumerateShows(int channels);
// void phoneUS.EnumerateHostsForShow(int channels, string showName);
// void phoneUS.AttachAsTalent(int channels, sting hostName);
// void phoneUS.ChangeShow(int channels, string showName, string showPass);
// void phoneUS.ConnectToShow(int channels, string showName, string showPass, string showHost);
// void phoneUS.EnumerateDirectors(int channels);
// void phoneUS.AttachToDirector(int channels, string directorName);
// void phoneUS.AttachAsTalent(int channels);
// live operations
// bool phoneUS.PostBusyAllButton(int channel);
// bool phoneUS.PostDelayDumpButton(int channel, bool down);
// bool phoneUS.PostDialKey(int channel, int key);
// bool phoneUS.PostDropButton(int channel, int column);
// bool phoneUS.PostHoldButton(int channel, int column);
// bool phoneUS.PostLineButton(int channel, int line, int column);
// bool phoneUS.PostMuteButton(int channel);
// bool phoneUS.PostNextButton(int channel);
// bool phoneUS.PostRecordButton(int channel);
// bool phoneUS.PostSpeakerButton(int channel);
// bool phoneUS.PostTransferButton(int channel);
// bool phoneUS.PostUpdateAll(int channel);
// bool phoneUS.PostText(int channel, int textType, int line, string text);
//
// void phoneEU.OpenConnection(int channel, string host, string user, string pass, int line);
// void phoneEU.CloseConnection(int channel);
// void phoneEU.CloseAllConnections(void);
//
// void phoneEU.PostDropButton(int channel);
// void phoneEU.PostHoldButton(int channel);
// void phoneEU.PostResumeButton(int channel);
// void phoneEU.PostLineButton(int channel);
// void phoneEU.PostDialKey(int channel, int key);
// void phoneEU.CallNumber(int channel, string number);
// void phoneEU.PostUpdateAll(int channel);
//

// ------------------------- * USphone connection management system * -------------------------

/*
phUS_ConnInfo[] stores connection information for each phone channel (module)

each item can be null in case channel is inactive
or Object
{
	state,   // {idle, conhost, conshow, attach, working}

// some special states in case we need to connect to inactive show:
/// shwlist  - getting list of shows to find out active
/// conashow - connecting to active show
/// chshow   - changing show to the requested one
// additional variable:
/// ashname  - active show name

	host, login, pass,				// box connection params
	shname, shpass, shhost,		// show parameters
	lnmode,		// { 12 | 24 | Auto }
	lnrev,		// line reverse { yes | no } // whether Hybrid1 is on the right
}

*/
const phUS_maxnum=15;	// total number of US phone channels
var phUS_ConnInfo=[];
var phUS_muteRinger=false;

/*
** phUS_UpdateChnInfo
**
** called when USphone channel configuration is changed
** we need to:
** - close all active phone connectons
** - clean all modules from leftover icons
** - parse new phone configuration; format is: "<host>,<user>,<pass>"
** - send new configuration to ACL
** - wait for further instructions from ACL's profile about connection parameters
*/
function phUS_UpdateChnInfo(ph_config)
{
	var phmsg={id:'WRNI'};
	var n, chn, line;
	var params;

// populate connection list
	for(n=0; n<phUS_maxnum; n++)
	{
		chn=n+1;
	/* delete previous connection information */
		if(chn in phUS_ConnInfo)
		{
		/* clear reconnect timeout if any */
			if('retry' in phUS_ConnInfo[chn])
			{
				host.clearTimeout(phUS_ConnInfo[chn].retry);
				delete phUS_ConnInfo[chn].retry;
			}
			delete phUS_ConnInfo[chn];
		}
	/* close connection */
		phoneUS.CloseConnection(chn);

		if(!(n in ph_config) || ph_config[n]==null) continue;

	// channel is 'captured' (enabled)
	// -- clear icons
		for(line=1; line<24; line++)
			mod.setPhoneIcon(chn, line, 0xFF);

	// -- get IP connection parameters
		params=ph_config[n].split(",");
		if(params.length<3)
		{
		// channel captured, but not configured; clear 1st icon too and forget about it
			mod.setPhoneIcon(chn, 0, 0xFF);
			continue;
		}

	// channel captured and configured
		phUS_ConnInfo[chn]={state: 'idle', host: params[0], login: params[1], pass: params[2]};
		mod.setPhoneIcon(chn, 0, icons.RectSpin); // set connection in progress icon
		phmsg[host.sprintf("sCH%02d", chn)]=ph_config[n];

		log.info_f("DBG: Phone channel %d: '%s', '%s', '%s'", chn, phUS_ConnInfo[chn].host, phUS_ConnInfo[chn].login, phUS_ConnInfo[chn].pass);
	}

// close all phone windows here, and reset all USphone variables
	resetGlobalPhoneData();

	return phmsg;
}

/*
** phUS_ShowConnect
**
** connect to phone show
*/
function phUS_ShowConnect(chn)
{
	var line;
	var phUS;

	if(!(chn in phUS_ConnInfo)) return;
	phUS=phUS_ConnInfo[chn];
	if(phUS.nshname=="")
	{
	// empty show name means disconnect
		phUS.shname="";
		phoneUS.CloseConnection(chn);
	// clear icons
		for(line=0; line<24; line++)
			mod.setPhoneIcon(chn, line, 0xFF);
		phUS.state='idle';
		initPhone(chn, 0, 0, false, false);
		return;
	}

/* copy new parameters over */
	phUS.shname=phUS.nshname;
	phUS.shpass=phUS.nshpass;
	phUS.shhost=phUS.nshhost;
	phUS.lnmode=phUS.nlnmode;
	phUS.lnrev =phUS.nlnrev;

	log.info_f("DBG: Phone %d: free to connect to '%s'", chn, phUS.shname);
	phUS.nshname="";

	mod.setPhoneIcon(chn, 0, icons.RectSpin); // set connection in progress icon
	for(line=1; line<24; line++)
		mod.setPhoneIcon(chn, line, 0xFF);
	phUS.state='conhost';
// opening TCP connection to Phone Box, awaiting phConnected
	phoneUS.OpenConnection(chn, phUS.host, phUS.login, phUS.pass);
}

/*
** phUS_UpdateShowInfo
**
** called when new show info is received from ACL
**
** cmsg sample: {id:"NEST",	mCH01:{id:"INDI", sPSWN:"Hybrid1&2", sHOST:"TWOx12", sPSWD:"", bHYBM:0, bPMOD:0}};
*/
function phUS_UpdateShowInfo(cmsg)
{
	var line, chn;
	var sh_config, params;
	var phUS;

	for(chn=1; chn<=phUS_maxnum; chn++)
	{
		if(!(chn in phUS_ConnInfo)) continue; // no such channel in capture profile
		phUS=phUS_ConnInfo[chn];
	/* clear reconnect timeout if any */
		if('retry' in phUS)
		{
			host.clearTimeout(phUS.retry);
			delete phUS.retry;
		}

		sh_config=(txtChn[chn] in cmsg)?cmsg[txtChn[chn]]:null;
		if(sh_config==null
		|| !('sPSWN' in sh_config) || !('sPSWD' in sh_config) || !('sHOST' in sh_config)
		|| !('bPMOD' in sh_config) || !('bHYBM' in sh_config))
		{
		// no data was received from ACL about this channel
		// if connection exists we should close it
			if(phUS.state=='idle') continue;
			if(phUS.state!='working' || monitorPhoneFaders(chn)==0)
			{
				phoneUS.CloseConnection(chn);
			// clear icons
				for(line=0; line<24; line++)
					mod.setPhoneIcon(chn, line, 0xFF);
				phUS.state='idle';
				initPhone(chn, 0, 0, false, false);
			}
		// in working state, and not eligible to close connection
			phUS.nshname="";
			continue;
		}

	// grab all the parameters we need
		phUS.nshname=sh_config.sPSWN;
		phUS.nshpass=sh_config.sPSWD;
		phUS.nshhost=sh_config.sHOST;
		phUS.nlnmode=(sh_config.bPMOD==0)?'Auto':(sh_config.bPMOD==1?'12':'24');
		phUS.nlnrev =(sh_config.bHYBM!=0)?'on':'off';

		log.info_f("DBG: Phone show %d: '%s', '%s', '%s', '%s', '%s'", chn, phUS.nshname, phUS.nshpass, phUS.nshhost, phUS.nlnmode, phUS.nlnrev);
		if(phUS.state=='working' && monitorPhoneFaders(chn)>0)
			continue; // cann't connect to show just yet, wait till all faders go off
		phUS_ShowConnect(chn);
	}
}

/*
** phConnected
**
** called from C code when connection to host is established
*/
function phConnected(chn)
{
	if(!(chn in phUS_ConnInfo))
	{
		if(chn in nx_data)
		{
			var nx=nx_data[chn];

			if(nx.state!='conhost') return;
			log.info_f("DBG: NXPhone: Connected to '%s', discovering active show", nx.host);
			nx.shname=null;
			nx.state='shwlist';
			nx.sh_skipped=0;
			phoneUS.EnumerateShows(chn);
		}
		return;
	}

	var phUS=phUS_ConnInfo[chn];

	if(phUS.state!='conhost') return;

	log.info_f("DBG: USPhone %d: Connected to '%s', connecting to show '%s'", chn, phUS.host, phUS.shname);
	phUS.state='conshow';
	phoneUS.ConnectToShow(chn, phUS.shname, phUS.shpass, phUS.shhost);
// opening connection to Show, awaiting showInfo, or error
}

/*
** phDisconnected
**
** called from C code when connection to host is broken (by timeout, by device or by our request)
*/
function phDisconnected(chn)
{
	var n;

	if(!(chn in phUS_ConnInfo))
	{
		if(chn in nx_data)
		{
			var nx=nx_data[chn];

			if(nx.state=='idle') return;
			log.info_f("DBG: NXPhone: Disconnected");
			nx.state='conhost';
			for(n in nx.hybrids)
			{
				updateChnIcon(nx.hybrids[n].fader); // set usual icon
				mod.setChnKeyLed(nx.hybrids[n].fader, keys.ChnSet, 0);
			}
		}
		return;
	}

	var phUS=phUS_ConnInfo[chn];

	if(phUS.state=='idle') return;
	log.info_f("DBG: USPhone %d: Disconnected", chn);

// set line icons, awaiting phConnected
	phUS.state='conhost';
	mod.setPhoneIcon(chn, 0, icons.RectSpin); // set connection in progress icon
	for(n=1; n<24; n++)
		mod.setPhoneIcon(chn, n, 0xFF);
}

/*
** phSendShowInfo
**
** phone box is sending us information about the show we are connected to
*/
function phSendShowInfo(chn, maxLines, maxHybrids)
{
	if(!(chn in phUS_ConnInfo))
	{
		if(chn in nx_data)
		{
			var nx=nx_data[chn];
			log.info_f("DBG: NXPhone: phone info: %d, %d", maxLines, maxHybrids);
		}
		return;
	}

	var phUS=phUS_ConnInfo[chn];

	if(phUS.state=='idle') return;
	log.info_f("DBG: USPhone %d: %d hybrids, %d lines", chn, maxHybrids, maxLines);

	if(maxHybrids<0) maxHybrids=0;
	phUS.numHybrids=maxHybrids;

	switch(phUS.lnmode)
	{
	case '12':
		phUS.numLines=12;
		break;
	case '24':
		phUS.numLines=24;
		break;
	case 'Auto':
	default:
		phUS.numLines=(maxLines>12)?24:12;
		break;
	}

	if(phUS.lnrev=='on')
		phUS.reversed=true;
	else
		phUS.reversed=false;

	log.info_f("DBG: USPhone %d: mode %d lines, %s", chn, phUS.numLines, phUS.reversed?"reversed":"normal");
}

/*
** phRetryConnection
*/
function phRetryConnection()
{
	var chn=this.chn;
	var phUS;

	if(!(chn in phUS_ConnInfo)) return;
	phUS=phUS_ConnInfo[chn];
	if(phUS.state!='retry' || !('retry' in phUS)) return;
	delete phUS.retry;

	mod.setPhoneIcon(chn, 0, icons.RectSpin); // set connection in progress icon
	phUS.state='conshow';
	phoneUS.ConnectToShow(chn, phUS.shname, phUS.shpass, phUS.shhost);
}

/*
** phAbortConnection
**
** during connection to phone server there was an error, abort
*/
function phAbortConnection(chn, reason)
{
	log.error_f("USPhone %d: %s", chn, reason);
	phUS_ConnInfo[chn].state='retry';
	mod.setPhoneIcon(chn, 0, icons.CrossBig); // show error icon
	phUS_ConnInfo[chn].retry=host.setTimeout(20000, phRetryConnection, {chn:chn});
}

function phNxAbortConnection(chn, reason)
{
	if(chn in nx_data)
	{
		var nx=nx_data[chn];

		log.error_f("NXPhone: %s", reason);
		nx.state='idle';
		phoneUS.CloseConnection(chn);
	}
}

/*
** phSendError
**
** box's response to our command
** despite it's name its called in case of success too
*/
function phSendError(chn, errorNumber)
{
	var n;

	if(errorNumber==99001) return;	// ping error response

	if(!(chn in phUS_ConnInfo))
	{
		if(chn in nx_data)
		{
			var nx=nx_data[chn];

			switch(nx.state)
			{
			case 'idle':
				return;
			case 'shwlist':
				if(errorNumber==2001)
				{
					phNxAbortConnection(chn, "Error getting show list");
					return;
				}
				break;
			case 'conshow':
				if(errorNumber==5)
				{
					log.info_f("DBG: NXPhone: Connected to active show, attaching as talent");
					nx.state='attach';
					phoneUS.AttachAsTalent(chn);
					return;
				}
				if(errorNumber==5001 || errorNumber==5002)
				{
					phNxAbortConnection(chn, "Error connecting to active show");
					return;
				}
				break;
			case 'attach':
				if(errorNumber==8001)
				{
					phNxAbortConnection(chn, "Error attaching as talent: Not connected to show");
					return;
				}
				if(errorNumber==8002)
				{
					phNxAbortConnection(chn, "Error attaching as talent: Attach failed");
					return;
				}
				if(errorNumber==8)
				{
					log.info_f("NXPhone: Attached as talent, working");
					nx.state='working';
					if(phUS_muteRinger) phoneUS.MuteRinger(chn, true);
					phoneUS.PostUpdateAll(chn);
					return;
				}
				break;
			}
		}
		return;
	}

	var phUS=phUS_ConnInfo[chn];

	switch(phUS.state)
	{
	case 'idle':
		return;
	case 'shwlist':
		if(errorNumber==2001)
		{
			phAbortConnection(chn, "Error getting show list");
			return;
		}
		break;
	case 'chshow':
		if(errorNumber==4)
		{
			log.info_f("DBG: USPhone %d: Connected to show, attaching as talent", chn);
			phUS.state='attach';
			phoneUS.AttachAsTalent(chn);
			return;
		}
		if(errorNumber==4001 || errorNumber==4002)
		{
			phAbortConnection(chn, "Error changing show");
			return;
		}
		break;
	case 'conashow':
		if(errorNumber==5)
		{
			log.info_f("DBG: USPhone %d: Connected to active show, changing show", chn);
			phUS.state='chshow';
			phoneUS.ChangeShow(chn, phUS.shname, phUS.shpass);
			return;
		}
		if(errorNumber==5001 || errorNumber==5002)
		{
			phAbortConnection(chn, "Error connecting to active show");
			return;
		}
		break;
	case 'conshow':
		if(errorNumber==5001)
		{
			log.warning_f("USPhone %d: Error connecting to show: Unable to find active show", chn);
		// not everything is lost though, we need to make a loop to active show first
			phUS.ashname=null;
			phUS.state='shwlist';
			phoneUS.EnumerateShows(chn);
			return;
		}
		if(errorNumber==5002)
		{
			phAbortConnection(chn, "Error connecting to show: Invalid show password");
			return;
		}
		if(errorNumber==5)
		{
			log.info_f("DBG: USPhone %d: Connected to Show, attaching as talent", chn);
			phUS.state='attach';
			phoneUS.AttachAsTalent(chn);
			return;
		}
		break;
	case 'attach':
		if(errorNumber==8001)
		{
			phAbortConnection(chn, "Error attaching as talent: Not connected to show");
			return;
		}
		if(errorNumber==8002)
		{
			phAbortConnection(chn, "Error attaching as talent: Attach failed");
			return;
		}
		if(errorNumber==8)
		{
			log.info_f("DBG: USPhone %d: Attached as talent, working", chn);
			phUS.state='working';
			if(phUS_muteRinger) phoneUS.MuteRinger(chn, true);
			phoneUS.PostUpdateAll(chn);

		// set line icons
			if(phUS.numLines==12)
			{
				for(n=0; n<12; n++)
				{
					mod.setPhoneIcon(chn, n, 0);
					mod.setPhoneIcon(chn, n+12, 0xFF);
				}
			}
			else
			{
				for(n=0; n<24; n++)
					mod.setPhoneIcon(chn, n, 0);
			}

		// now initialize the hierarchical data structs associated w/ this phone module
			initPhone(chn, phUS.numLines, phUS.numHybrids, phUS.numLines>12?true:false, !phUS.reversed);
			return;
		}
		break;
	}

//log.info_f("Phone [%2d]: SendError(%d)", chn, errorNumber);
}

function phEnumerateShowsCallback(chn, showName, rqPass, isActive, hostName, isLast)
{
//  log.info_f("DBG: USPhone %d: enumerateshows(%s, %d, %d, %s, %d)", chn, showName, rqPass, isActive, hostName, isLast);
	if(!(chn in phUS_ConnInfo))
	{
		if(chn in nx_data)
		{
			var nx=nx_data[chn];

			if(nx.state!='shwlist') return;

		// remember active show
			if(isActive && !rqPass)
			{
				if(nx.shname==null || nx.sh_skipped<=nx.sh_skip)
				{
					nx.shname=showName;
					nx.hostName=hostName;
				}
				nx.sh_skipped++;
			}

			if(isLast)
			{
				if(nx.shname==null)
				{
				/* active show not found, just close connection */
					phNxAbortConnection(chn, "Unable to find active show");
					return;
				}
				log.info_f("DBG: NXPhone: Connecting to active show '%s@%s'", nx.shname, nx.hostName);
				nx.state='conshow';
				phoneUS.ConnectToShow(nx.chn, nx.shname, "", nx.hostName);
			}
		}
		return;
	}

	var phUS=phUS_ConnInfo[chn];

	if(phUS.state!='shwlist') return;

// remember active show
	if(phUS.ashname==null && isActive && hostName==phUS.shhost)
		phUS.ashname=showName;

	if(isLast)
	{
		if(phUS.ashname==null)
		{
		/* active show not found, just close connection */
			phAbortConnection(chn, "Unable to find active show @ host "+phUS.shhost);
			return;
		}
		log.info_f("DBG: USPhone %d: Connecting to active show '%s'", chn, phUS.ashname);
		phUS.state='conashow';
		phoneUS.ConnectToShow(chn, phUS.ashname, phUS.shpass, phUS.shhost);
	}
}

function phEnumerateShowsForHostCallback(chn, showName, isLast) {}
function phEnumerateHostsCallback(chn, hostName, isLast) {}
function phEnumerateDirectorsCallback(chn, directorName, isTalent, isFree, isLast) {}

// ------- Phone functions not related to a particular phone box type ----------------------------

function resetHoldKeyLed()
{
	holdkeyledTO=host.clearTimeout(holdkeyledTO);
	mod.setChnKeyLed(holdkeyled_fader, keys.ChnHold, 0);
	holdkeyled_fader=0;
}

function turnonHoldKeyLed(fader)
{
	if(holdkeyledTO!=null) resetHoldKeyLed();

	holdkeyled_fader=fader;
	holdkeyledTO=host.setTimeout(holdkeyledTOval, resetHoldKeyLed);
	mod.setChnKeyLed(fader, keys.ChnHold, 0xff);
}

// ------- Phone functions related to the so-called "US-Phone" boxen (e.g., TWOx12, 2101) -------

/* This software, which operates in the limits of one "console" (i.e., Element)
 * uses the following assumptions about call-controller (AKA us-phone) modules,
 * phone-boxes (e.g., 2101 or TWOx12) with their hybrids and lines, and faders
 * present on the console.
 *
 * There can be multiple CC-modules present in console, and multiple phone-boxes
 * on the network. Each CC-module maintains a connection to some phone-box, and,
 * furthermore, it has selected some specific "phone_show" configured for this
 * phone_box. Multiple CC-modules may be connected to the same phone-box, but not
 * to the same "phone-show". A "phone_id" AKA "phone channel" is the software
 * identifier assigned to each CC-module.
 *
 * The non-fader ("middle") buttons of each CC-module can only work with this module,
 * i.e., they are permanently associated with their CC-module's phone_id and cannot be
 * made a part of a different CC-module. Two configurable boolean parameters of each
 * CC-module affect the operation of its middle buttons: "mode24" makes both
 * columns of line-selection buttons work in tandem and generally apply to
 * "Hybrid1" (hybrid with index 0 for the purposes of communicating to the
 * phone box). When "mode24" is not set, the operation of the left and right
 * line-selection button columns differs and the "hyb1left" parameter becomes
 * significant. It controls whether the left or right column of line-selection
 * buttons applies to "Hybrid1". It also determines whether the left or the right
 * arrow should be displayed for On-Air lines, and this role is important
 * even in 24-line mode. The "Next" button in a CC-module always
 * works with "Hybrid1". The other hybrids are called "sequenceable".
 *
 * The faders physically attached to each CC-module are completely independent
 * of the middle-buttons and share the same freedom of association to hybrids
 * and lines as any other fader. They don't even have to be associated with any
 * CC-module and can be used as non-phone faders. In general, for each fader
 * a source is picked. Some sources are associated with CC-modules. In this case,
 * their configuration information carries the following extra attributes:
 * "phone_id", "hybrid", "mashing" and "fixedline". Please note that the
 * same source can be picked for more than one fader. Please also note that
 * different sources can refer to the same hybrid. The "phone_id" attribute
 * of the source ties the fader to a particular CC-module. The "hybrid"
 * attribute specifies the hybrid ID (unique within one phone-show and thus
 * per phone_id) that this fader will work with. Index 0 is reserved for "Hybrid1",
 * whose role is somewhat special, as described above. The "fixedline"
 * attribute allows tying a source (and thus the faders that have selected it)
 * to a particular line. The "mashing" attribute specifies whether more than
 * one line can be conferenced together on the hybrid.
 */

var holdkeyledTOval = 500;
var holdkeyledTO = null;
var holdkeyled_fader = 0;

// Last used line is used in conjunction with the global Last Caller and Dialout popups.
// However, since line IDs can be the same on different phone_id-s, we need to also
// remember the "last phone_id" which caused the popup to be displayed.
var last_used_line =
	{
		line_id  : -1, // impossible
		phone_id : -1  // impossible
	};

var takelineTOval = 10000; // "Take Line" timeout, in ms

var phone_mod_data = {}; // for each phone_id, holds a struct with lots of interesting info and data fields

// Note that each ch_data[i] has a field us_phone that is a struct holding phone_id, hybrid, mashing and fixed_line fields

function resetGlobalPhoneData()
{
	us_DialoutFrClose();
	us_LastcallFrClose();
	last_used_line.line_id = -1;
	last_used_line.phone_id = -1;
}

// a frequent test for whether a particular phone channel has been initialized (i.e. connected to a show, etc.)
function isPhoneInited(phone_id)
{
	return (phone_id in phone_mod_data) && phone_mod_data[phone_id] != null;
}

function initPhone(phone_id, numlines, numhybrids, mode24, hyb1left)
{
	var i;
//	log.info_f("initPhone: phone_id = %d, numlines = %d, numhybrids = %d, mode24 = %d, hyb1left = %s", phone_id, numlines, numhybrids, mode24, hyb1left ? "true" : "false");

	if (numlines == 0 || numhybrids == 0) // connecting to a show or something has failed, clear the data for this phone_id
	{
		phone_mod_data[phone_id] = null;

		for(i=1; i<=64; i++)     // clear the faders formerly associated with this phone channel
			if(ch_data[i].us_phone!=null && ch_data[i].us_phone.phone_id==phone_id)
				ch_data[i].us_phone=null;
	}
	else
	{
		phone_mod_data[phone_id] =
			{
				mode24               : mode24,     // module working in 24-line mode or not
				hyb1left             : hyb1left,   // Hybrid1 positioned to the left of module center or not
				numlines             : numlines,   // how many lines can be attached (12 or 24, likely)
				numhybrids           : numhybrids, // how many hybrids can be attached
				num_mon_faders       : 0,          // number of faders to monitor for "off" condition, see below
				takelinehybrid       : -1,         // which hybrid will take the next line assignment
				takelineTO           : null,       // the timeout object for the hybrid mentioned above
				last_used_seq_hybrid : -1,         // which of the sequenceable hybrids was last used
				line_data            : [],         // indexed by lines
				hybrid_data          : [],         // indexed by hybrids
				mon_faders           : {}          // faders to monitor for "off" condition, needed on show-change
			};

		for (i = 0; i < numlines; i++)
			phone_mod_data[phone_id].line_data[i] =
				{
					fixed_on_hyb  : -1,  // which hybrid this line is fixed on
					used_on_hyb   : -1,  // which hybrid this line is in use on
					locked        : false, // is it locked on the hybrid that it's in use on
					state         : 0    // what state this line was last seen in (0 = invalid)
				};

		for (i = 0; i < numhybrids; i++)
			phone_mod_data[phone_id].hybrid_data[i] =
				{
					fixed_on_line  : -1,   // which line this hybrid is fixed on
					mashing        : true, // whether mashing more than one line to this hybrid is OK
					assignedfaders : []    // faders associated with this hybrid (few elements, not 64)
				};

		// refresh the line and hybrid data that was just reset from any
		// info that could be present in ch_data[x].us_phone structs
		for (i = 1; i <= 64; i++)
		{
			var phone_info = ch_data[i].us_phone;

			if (phone_info != null && phone_info.phone_id == phone_id)
				updatePhoneData(phone_id, i, phone_info);
		}
	}

// once a phone-show has loaded, we can hide the on-screen warning
	gui_indi.setPhoneShowWaiting(false);
}

function updatePhoneData(phone_id, fader, phone_info)
{
	// sometimes we do get this call before initPhone, must ignore...
	// and we ignore it if the hybrid number exceeds what the box physically has connected
	if (!isPhoneInited(phone_id) || phone_info.hybrid >= phone_mod_data[phone_id].numhybrids)
		return;

//	log.info_f("updatePhoneData: phone_id = %d, fader = %d, hybrid = %d, mashing = %s, fixedline = %d", phone_id, fader, phone_info.hybrid, phone_info.mashing ? "true" : "false", phone_info.fixedline);

	phone_mod_data[phone_id].hybrid_data[phone_info.hybrid].fixed_on_line = phone_info.fixedline; // -1 or 0 to 23
	phone_mod_data[phone_id].hybrid_data[phone_info.hybrid].mashing = phone_info.mashing;

	var faders = phone_mod_data[phone_id].hybrid_data[phone_info.hybrid].assignedfaders;
	var found = false;

	for (var j in faders)
		if (faders[j] == fader)
		{
			found = true;
			break;
		}

	if (!found)
		phone_mod_data[phone_id].hybrid_data[phone_info.hybrid].assignedfaders.push(fader);

	if (phone_info.fixedline >= 0)
		phone_mod_data[phone_id].line_data[phone_info.fixedline].fixed_on_hyb = phone_info.hybrid;
}

function clearPhoneData(fader)
{
//	log.info_f("clearPhoneData: fader = %d", fader);

	for (var pi in phone_mod_data)
	{
		if (!isPhoneInited(pi))
			continue;

		for (var i = 0; i < phone_mod_data[pi].numhybrids; i++)
		{
			var faders = phone_mod_data[pi].hybrid_data[i].assignedfaders;

			for (var j in faders)
				if (faders[j] == fader)
				{
					faders.splice(j, 1); // (updates the length, too, as opposed to delete faders[j])
					break;
				}
		}
	}
}

function printPhoneData(phone_id)
{
	var i, j;
	log.info_f("printPhoneData(%d): mode24 = %s, hyb1left = %s, takelinehybrid = %d, last_used_seq_hybrid = %d",
		   phone_id, phone_mod_data[phone_id].mode24 ? "true" : "false",
		   phone_mod_data[phone_id].hyb1left ? "true" : "false",
		   phone_mod_data[phone_id].takelinehybrid, phone_mod_data[phone_id].last_used_seq_hybrid);

	for (i = 0; i < phone_mod_data[phone_id].numhybrids; i++)
	{
		var hybrid_data = phone_mod_data[phone_id].hybrid_data[i];
		var faders = hybrid_data.assignedfaders;
		var faderstr = "";

		for (j in faders)
			faderstr += " " + faders[j];

		log.info_f("hybrid %d info: mashing = %s, fixed_on_line = %d, assigned faders: %s",
			   i, hybrid_data.mashing ? "true" : "false", hybrid_data.fixed_on_line, faderstr);
	}

	for (i = 0; i < phone_mod_data[phone_id].numlines; i++)
	{
		var line_data = phone_mod_data[phone_id].line_data[i];
		log.info_f("line %d info: fixed_on_hyb = %d, used_on_hyb = %d, locked = %s, state = 0x%04X\n",
			   i, line_data.fixed_on_hyb, line_data.used_on_hyb,
			   line_data.locked ? "yes" : "no", line_data.state);
	}
}

// Called by host when phone shows are being changed (or initially set). It is a request to the USphone subsystem
// to start monitoring the faders (associated with the particular phone-module's hybrids) that are currently "on".
// Whenever such a fader goes "off" it can be dropped from the monitoring-list. When all have been dropped, a special
// callback is invoked. The function returns the number of faders to be monitored. If 0, no callback will be called.
function monitorPhoneFaders(phone_id)
{
	gui_indi.setPhoneShowWaiting(true);

	if (!isPhoneInited(phone_id)) // if there is no active phone-show
		return 0;             // then there can't be anything to monitor yet

	var takelinehybrid = phone_mod_data[phone_id].takelinehybrid;

	// get rid of any Take Line indications we may have, cancel Take Line timeout
	if (takelinehybrid >= 0)
	{
		phone_mod_data[phone_id].takelineTO=host.clearTimeout(phone_mod_data[phone_id].takelineTO);
		phone_mod_data[phone_id].takelinehybrid = -1;
		showText(phone_id, takelinehybrid);
	}

	phone_mod_data[phone_id].num_mon_faders=0;
	phone_mod_data[phone_id].mon_faders=[];
	var faders_to_monitor = "";

	for (var i = 1; i <= 64; i++)
	{
		var phone_info = ch_data[i].us_phone;

		// if this fader is associated with USphone-s and, in particular, with our module
		if (phone_info != null && phone_info.phone_id == phone_id
		    && ch_data[i].onsw) // and it happens to be "on"
		{
			phone_mod_data[phone_id].mon_faders[i]=true; // then remember it
			phone_mod_data[phone_id].num_mon_faders++;
			faders_to_monitor += " " + i;
		}
	}

	log.info_f("Monitor Phone Faders: phone_id = %d, num_mon_faders = %d, mon_faders = %s",
		   phone_id, phone_mod_data[phone_id].num_mon_faders, faders_to_monitor);
	return phone_mod_data[phone_id].num_mon_faders;
}

// called by host.js when a USphone fader goes off
function notifyPhoneFaderOff(phone_id, fader)
{
	if (!isPhoneInited(phone_id) // no active phone-show (it probably won't even come here if so)
	    || phone_mod_data[phone_id].num_mon_faders <= 0) // or nothing being monitored right now
		return;

	if((fader in phone_mod_data[phone_id].mon_faders) && phone_mod_data[phone_id].mon_faders[fader]) // if a fader we're monitoring goes off
	{
		phone_mod_data[phone_id].mon_faders[fader]=false;
		phone_mod_data[phone_id].num_mon_faders--;
		log.info_f("notifyPhoneFaderOff: phone_id = %d, fader %d dropped from list", phone_id, fader);
	}

	if(phone_mod_data[phone_id].num_mon_faders<=0) // if this was the last one we're monitoring
	{
		log.info_f("notifyPhoneFaderOff: phone_id = %d, no more monitorable faders", phone_id);
		phUS_ShowConnect(phone_id); // call the callback
	}
}

function onPhoneKeyDown(phone_id, key)
{
//	log.info_f("Phone [%2d]: key %2d pressed", phone_id, key);

	if(!isPhoneInited(phone_id)) // no active phone-show currently
		return;

	// Hybrid1 is not necessarily the one that we'll end up using, but for the purposes of
	// setting last_used_phone_ch (fader, really) it doesn't matter. The only reason we need
	// it is for host.js to be able to find the right phone module (with the right phone_id)
	// in order to send any numpad key presses. Therefore, it suffices to pretend we used Hybrid1.
	// One more complication... only relevant on an abnormal setup... Hybrid1 need not be in use...
	// So check all hybrids in order until we get one that has a fader assigned to it.
	for(var i = 0; i < phone_mod_data[phone_id].numhybrids; i++)
		if((last_used_phone_ch = getFirstFader(phone_id, i)) != null)
			break;

	if(key>=0 && key<12)
		processPhoneKey(phone_id, key, true); // true means "left side"
	else if(key>=12 && key<24)
	{
		var line=phone_mod_data[phone_id].mode24?key:key-12;
		processPhoneKey(phone_id, line, false); // false means "right side"
	}
	else switch(key)
	{
	case keys.PhnXfer:
		phoneUS.PostTransferButton(phone_id);
		break;
	case keys.PhnBusyAll:
		phoneUS.PostBusyAllButton(phone_id); // block all
		break;
	case keys.PhnDropLt:
		// in 24-line mode both Drop buttons affect Hybrid1 only, no way to drop sequenceable hybrids
		// in 12-line mode this button affects Hybrid1 if it's on left
		if(phone_mod_data[phone_id].mode24 || phone_mod_data[phone_id].hyb1left)
			phoneUS.PostDropButton(phone_id, 0); // 0 is the index for Hybrid1
		else // 12 line mode with Hybrid1 on right, so this button drops the last sequenceable hybrid
			phoneUS.PostDropButton(phone_id, phone_mod_data[phone_id].last_used_seq_hybrid);
		break;
	case keys.PhnDropRt:
		// in 24-line mode both Drop buttons affect Hybrid1 only, no way to drop sequenceable hybrids
		// in 12-line mode this button affects Hybrid1 if it's on right
		if(phone_mod_data[phone_id].mode24 || !phone_mod_data[phone_id].hyb1left)
			phoneUS.PostDropButton(phone_id, 0); // 0 is the index for Hybrid1
		else // 12 line mode with Hybrid1 on left, so this button drops the last sequenceable hybrid
			phoneUS.PostDropButton(phone_id, phone_mod_data[phone_id].last_used_seq_hybrid);
		break;
	case keys.PhnNext:
		phoneUS.PostNextButton(phone_id); // always applies to Hybrid1
		break;
	}
}

// converts a key press to tryPostLineButton, given what line it corresponds to and which side of module was it pressed
function processPhoneKey(phone_id, line, leftside)
{
	var fixed_hybrid = phone_mod_data[phone_id].line_data[line].fixed_on_hyb;

	if (fixed_hybrid >= 0) // this is a fixed line, no other hybrids even usable w/ it
		tryPostLineButton(phone_id, line, fixed_hybrid);
	else
	{
		var takelinehybrid = phone_mod_data[phone_id].takelinehybrid;

		if (takelinehybrid >= 0) // a dedicated hybrid has been selected, use it
			tryPostLineButton(phone_id, line, takelinehybrid);
		else
		{
			var usedonhybrid = phone_mod_data[phone_id].line_data[line].used_on_hyb;

			if (usedonhybrid >= 0 // if the line is already in use on a hybrid
			    && (phone_mod_data[phone_id].mode24 // and we're in 24-line mode
				// or it is one of the sequenceable hybrids that (probably) are on the same side as key-press
				|| (usedonhybrid > 0 && (phone_mod_data[phone_id].hyb1left != leftside))
				// or it is in use on Hybrid1 which happens to be on the same side as key-press
				|| (usedonhybrid == 0 && (phone_mod_data[phone_id].hyb1left == leftside))))
				tryPostLineButton(phone_id, line, usedonhybrid); // then apply it to this hybrid
			// it's an unused line or a line that should migrate:
			// 24 line mode, or 12 line mode with Hybrid1 on the same side as key-press
			else if (phone_mod_data[phone_id].mode24 || (phone_mod_data[phone_id].hyb1left == leftside))
				tryPostLineButton(phone_id, line, 0); // 0 is the index for Hybrid1
			else // otherwise we have to find the best hybrid for this call, considering busy and mashing
			{
				var best_hybrid = findBestHybrid(phone_id, line);

				if (best_hybrid >= 0) // found one that's free or where we could mash this call
					tryPostLineButton(phone_id, line, best_hybrid);
			}
		}
	}
}

function onPhoneKeyUp(phone_id, key)
{
//	log.info_f("Phone [%2d]: key %2d released", phone_id, key);
}

function onPhoneFaderKeyDown(phone_id, fader, key)
{
	if(phone_id==-1)
	{
	// Euro Phone case
		onEUphoneKeyDown(fader, key);
		return;
	}

//	log.info_f("Phone [%2d]: fader[%2d] key %2d pressed", phone_id, fader, key);

	if(!isPhoneInited(phone_id)) // no active phone-show currently
		return;

	var hybrid = ch_data[fader].us_phone.hybrid;

	// if no hybrid associated with this fader, then nothing else to do here
	if (hybrid < 0 || hybrid >= phone_mod_data[phone_id].numhybrids)
		return;

	var fixed_line = phone_mod_data[phone_id].hybrid_data[hybrid].fixed_on_line;

	switch(key)
	{
	case keys.ChnHold:
	// if it's a fixed line and happens to be ringing in right now, answer it first
		if (fixed_line >= 0 && phone_mod_data[phone_id].line_data[fixed_line].state == 0x0301)
		{
			tryPostLineButton(phone_id, fixed_line, hybrid); // the "try" is probably unnecessary
			host.sleep(100); // Telos2 did this, so we do it, too
		}
		phoneUS.PostHoldButton(phone_id, hybrid);
		break;
	case keys.ChnON:
		if(!ch_data[fader].onsw) break; // if the channel is OFF, the ON button should just turn it ON

		if(fixed_line >= 0) // for fixed lines react the same way as if the line button were pressed
			tryPostLineButton(phone_id, fixed_line, hybrid);
		// not a fixed line, but see whether it makes sense to create the Take Line condition
		else if (phone_mod_data[phone_id].hybrid_data[hybrid].mashing // mashing allowed on hybrid
			 || getFirstLockedLine(phone_id, hybrid) < 0)         // or no lines locked on it yet
			changeTakeLine(phone_id, hybrid);
		break;
	case keys.ChnOFF:
		if(ch_data[fader].onsw) break; // if the channel is ON, the OFF button should just turn it OFF

		// If it is OFF already, however, then it should drop any lines that are droppable on it.
		// If and only if there are no droppable lines, it should cause the "Take Line" effect.
		// So start guessing whether there are droppable lines on this hybrid...
		var i;

		for (i = 0; i < phone_mod_data[phone_id].numlines; i++)
			if (phone_mod_data[phone_id].line_data[i].used_on_hyb == hybrid // a line is used here
			    && !phone_mod_data[phone_id].line_data[i].locked) // and it ain't locked
				break;

		if (i < phone_mod_data[phone_id].numlines)        // found a line that should be droppable
			phoneUS.PostDropButton(phone_id, hybrid); // then trying to drop it is all we do
		// but if nothing seems droppable, consider making it the "Take Line" hybrid, even for fixed ones
		else if (phone_mod_data[phone_id].hybrid_data[hybrid].mashing // mashing allowed on hybrid
			 || getFirstLockedLine(phone_id, hybrid) < 0)         // or no lines locked on it yet
			changeTakeLine(phone_id, hybrid);
		break;
	case keys.ChnPreview:
		// FALLTHROUGH
	default:
		if (fixed_line >= 0) // for fixed lines react the same way as if the line button were pressed
			tryPostLineButton(phone_id, fixed_line, hybrid);
		// but for the rest check whether it makes sense to create the Take Line condition there
		else if (phone_mod_data[phone_id].hybrid_data[hybrid].mashing // mashing allowed on hybrid
			 || getFirstLockedLine(phone_id, hybrid) < 0)         // or no lines locked on it yet
			changeTakeLine(phone_id, hybrid);
	}
}

/*
** onPhoneDTMFkeyDown
**
** called when DTMF keypad is free and key is pressed
** key: 0..11 (0-9: '0'..'9', 10: '*', 11: '#')
*/
var phDTMFtoSym=[48,49,50,51,52,53,54,55,56,57,42,35];
function onPhoneDTMFkeyDown(phone_id, key)
{
	var sym=phDTMFtoSym[key];
//	log.info_f("Phone [%2d]: DTMF key %2d '%c' pressed", phone_id, key, sym);

	if(!isPhoneInited(phone_id)) return; // this probably ain't possible now, but just in case we test it

	phoneUS.PostDialKey(phone_id, sym);

// animation
	gui_indi.animUSdial(key==0?10:(key==11?key:key-1));
}

function onPhoneDTMFkeyUp(phone_id, key)
{
//	log.info_f("Phone [%2d]: DTMF key %2d '%c' released", phone_id, key, phDTMFtoSym[key]);
	gui_indi.animUSdial(-1);
}

// ------------------------- * callback functions * -------------------------

function phSendMode(phone_id, talent)
{
//	log.info_f("Phone [%2d]: SendMode(%d, %d)", phone_id, talent);
}

function phSendShowChange(chn, newShowName, oldShowName)
{
	if(!(chn in phUS_ConnInfo)) return; // not a US phone
	var line, phUS=phUS_ConnInfo[chn];

	if(phUS.state!='working') return;

	log.info_f("DBG: USPhone %d: Show changed from '%s' to '%s'", chn, oldShowName, newShowName);
	if(newShowName=="<show_disconnected>")
	{
		mod.setPhoneIcon(chn, 0, icons.RectSpin); // set connection in progress icon
		for(line=1; line<24; line++)
			mod.setPhoneIcon(chn, line, 0xFF);
		phUS.state='conshow';
		phoneUS.ConnectToShow(chn, phUS.shname, phUS.shpass, phUS.shhost);
	// opening connection to Show, awaiting showInfo, or error
	}
	else // update show name, use it in case of connection loss
		phUS.shname=newShowName;
}

function phSendRecord(phone_id, on)
{
//	log.info_f("Phone [%2d]: SendRecord(%d)", phone_id, on);
}

function phSendBusyAll(phone_id, on)
{
//	log.info_f("Phone [%2d]: SendBusyAll(%d)", phone_id, on);
}

function phSendTransfer(phone_id, on)
{
//	log.info_f("Phone [%2d]: SendTransfer(%d)", phone_id, on);
}

function phSendMute(phone_id, on)
{
//	log.info_f("Phone [%2d]: SendMute(%d)", phone_id, on);
}

function phSendRingMute(phone_id, on)
{
//	log.info_f("Phone [%2d]: SendRingMute(%d)", phone_id, on);
}

function phSendDelayDump(phone_id, on)
{
//	log.info_f("Phone [%2d]: SendDelayDump(%d)", phone_id, on);
}

function phSendHandset(phone_id, on)
{
//	log.info_f("Phone [%2d]: SendHandset(%d)", phone_id, on);
}

function phSendSpeaker(phone_id, on)
{
//	log.info_f("Phone [%2d]: SendSpeaker(%d)", phone_id, on);
}

function phSendLineState(phone_id, line, hybrid, lineState, elapsedTimeSec)
{
	var oldhybrid;
	var icon;
	var dir_left;

	if(!(phone_id in phUS_ConnInfo))
	{
		if(phone_id in nx_data)
		{
			var nx=nx_data[phone_id];
			var n, h, cnx;
			var set_style, drop;

			if(nx.state!='working') return;
			if(!(line in nx.lines)) return;	// not our line

			n=nx.lines[line];	// hybrid number
			if(n!=hybrid)
			{
				if((n>>1)!=(hybrid>>1)) return;
			// hybrid information is paired
				switch(lineState)
				{
				case phUSstate.invalid:
				case phUSstate.reserved:
				case phUSstate.idle:
				case phUSstate.ringing:
				case phUSstate.hold:
				case phUSstate.talentHold:
				case phUSstate.readyHold:
				case phUSstate.readyNext:
				case phUSstate.busyAll:
				case phUSstate.usedElsewhere:
				case phUSstate.sharedUnavailable:
				// hybrid-neutral states
					hybrid=n;
					break;
				case phUSstate.onAir:
				case phUSstate.onAirLocked:
				case phUSstate.handset:
				case phUSstate.handsetLocked:
				case phUSstate.speaker:
				case phUSstate.speakerLocked:
				// other hybrid states
					hybrid=n;
					lineState=phUSstate.usedElsewhere;
					break;
				default:
					return;	// ignore all other states
				}
			}
			if(!(n in nx.hybrids)) return;		// not supposed to happen
			h=nx.hybrids[n];
			cnx=ch_data[h.fader].nx_phone;
			if(cnx==null) return;
			cnx.line_state=lineState;

		//	log.info_f("DBG: NX state: line %d, on hybrid %d, state: %d", line, hybrid, lineState);

		/*
			here we should prevent others from stealing both our lines and our hybrids
			1. if it's our line, but it's used on other hybrid, bring it back to our hybrid using command:
			phoneUS.PostLineButton(cnx.chn, cnx.line, cnx.hybrid);
			2. ??? if other line is used on our hybrid then set it on hold, but only if our line is not OnAir otherwise we will lose our line
			phoneUS.PostHoldButton(cnx.chn, cnx.hybrid);
		*/

			icon=0; set_style=0; drop=false;
			switch(lineState)
			{
			default:
			case phUSstate.invalid:
				break;
			case phUSstate.reserved: // Reserved [dot] (ready)
			case phUSstate.idle: // Idle [dot]
				icon=icons.Dot;
				drop=true;
				break;
			case phUSstate.ringing: // Ringing [box animation] (incoming)
				icon=icons.Ringing;
				break;
			case phUSstate.hold: // Hold [box in frame animation]
			case phUSstate.talentHold: // Talent Hold [?]
				icon=icons.OnHold;
				drop=true;
				break;
			case phUSstate.readyHold: // Ready Hold [check]
			case phUSstate.readyNext: // Ready Next [check/no-check animation]
				icon=icons.CheckBig;
				break;
			case phUSstate.onAir: // On Air [Arrow] (seized)
				icon=icons.ArrowDn;
				set_style=255;
				break;
			case phUSstate.onAirLocked: // On Air Locked [lock w/ triangle]
				icon=icons.LockLt;
				break;
			// FIXME the following 4 cases are very questionable, but most likely not in use yet
			case phUSstate.handset: // Handset [handset]
			case phUSstate.handsetLocked: // Handset Locked [?]
				icon=icons.Handset;
				break;
			case phUSstate.speaker: // Speaker [?]
			case phUSstate.speakerLocked: // Speaker Locked [?]
				break;
			case phUSstate.usedElsewhere: // Used Elsewhere [small cross]
				icon=icons.CrossSml;
				break;
			case phUSstate.busyAll: // Busy All [big cross]
				icon=icons.CrossBig;
				break;
			case phUSstate.sharedUnavailable: // Shared Unavailable [arrow pointing right]
				icon=icons.CrossSml;
				break;
			}
			if(icon>=0) mod.setChnIcon(h.fader, icon, elapsedTimeSec>0?1:0);
			mod.setChnKeyLed(h.fader, keys.ChnSet, set_style);
			if(drop) chnLineDroped(h.fader);
			if(h.fader==ringing_fr_channel && lineState!=phUSstate.ringing)
			{
			// the channel for which we last displayed ringing_fr ain't ringin' no more
				ringingTO=host.clearTimeout(ringingTO); // there must be a timeout related to it
				ringingFrClose();
			}
		}
		return;
	}

	var phUS=phUS_ConnInfo[phone_id];

// allegedly, this can happen in TWOx12, at least the != 'working' case...
	if(phUS.state!='working')
		return;

// ignore lines with which we're not supposed to work with
	if(line>=phone_mod_data[phone_id].numlines) return;
// ignore hybrids which we were told don't exist (somebody may be adding them on...)
	if(hybrid>=phone_mod_data[phone_id].numhybrids) return;

// elapsedTimeSec really means "how long the line has been in this state"
//	log.info_f("Phone [%2d]: SendLineState(%d, %d, 0x%X, %d)", phone_id, line, hybrid, lineState, elapsedTimeSec);

// assume a line in use is becoming (or staying unlocked), correct later for the rare cases that this ain't so
	if(phone_mod_data[phone_id].line_data[line].used_on_hyb==hybrid) // line state could be anything on other hybrids
		phone_mod_data[phone_id].line_data[line].locked=false;

	icon=255; // empty? well, it's something we want displayed in most bad cases...

// icon direction
	if(phone_mod_data[phone_id].hyb1left)
		dir_left=(hybrid==0); // if Hybrid1 is to the left  of CC-module
	else
		dir_left=(hybrid!=0); // if Hybrid1 is to the right of CC-module

	switch(lineState)
	{
	case phUSstate.invalid:
		if(line==last_used_line.line_id && phone_id==last_used_line.phone_id)
			us_InitiatePhoneFrClose();
		break;
	case phUSstate.reserved: // Reserved [dot] (ready)
	case phUSstate.idle: // Idle [dot]
		if(line==last_used_line.line_id && phone_id==last_used_line.phone_id)
			us_InitiatePhoneFrClose();
		icon=icons.Dot;
	// we can't depend on phSendDropLine always being called, so do it ourselves
		markLineAsFree(phone_id, line);
		break;
	case phUSstate.ringing: // Ringing [box animation] (incoming)
		icon=icons.Ringing;
		break;
	case phUSstate.hold: // Hold [box in frame animation]
		icon=icons.OnHold;
		markLineAsFree(phone_id, line); // hold is a line-specific condition, so free the hybrid
		break;
	case phUSstate.readyHold: // Ready Hold [check]
		icon=icons.CheckBig;
		markLineAsFree(phone_id, line); // hold is a line-specific condition, so free the hybrid
		break;
	case phUSstate.readyNext: // Ready Next [check/no-check animation]
		icon=icons.CheckBig; // ?
		markLineAsFree(phone_id, line); // hold is a line-specific condition, so free the hybrid
		break;
	case phUSstate.talentHold: // Talent Hold [?]
		icon=icons.OnHold; // ?
		markLineAsFree(phone_id, line); // hold is a line-specific condition, so free the hybrid
		break;
	case phUSstate.onAir: // On Air [Arrow] (seized)
		icon=dir_left?icons.ArrowLt:icons.ArrowRt;
		oldhybrid=phone_mod_data[phone_id].line_data[line].used_on_hyb;

	// when there is a change in hybrids, no phSendSeizeLine or phSendSelectLine
	// gets called, so make sure that we remember where this line is now used...
	// FIXME? Other states that need this?
		if(oldhybrid!=hybrid)
		{
			markLineAsUsed(phone_id, line, hybrid);
			showText(phone_id, oldhybrid); // also update the display for the old hybrid
		}
		break;
	case phUSstate.onAirLocked: // On Air Locked [lock w/ triangle]
		icon=dir_left?icons.LockLt:icons.LockRt;
		phone_mod_data[phone_id].line_data[line].locked=true;
		oldhybrid=phone_mod_data[phone_id].line_data[line].used_on_hyb;

	// when there is a change in hybrids, no phSendSeizeLine or phSendSelectLine
	// gets called, so make sure that we remember where this line is now used...
	// FIXME? Other states that need this?
		if(oldhybrid!=hybrid)
		{
			markLineAsUsed(phone_id, line, hybrid);
			showText(phone_id, oldhybrid); // also update the display for the old hybrid
		}
		break;

	// FIXME the following 4 cases are very questionable, but most likely not in use yet
	case phUSstate.handset: // Handset [handset]
		icon=icons.Handset;
		break;
	case phUSstate.handsetLocked: // Handset Locked [?]
		icon=icons.Handset;
		phone_mod_data[phone_id].line_data[line].locked=true;
		break;
	case phUSstate.speaker: // Speaker [?]
		break;
	case phUSstate.speakerLocked: // Speaker Locked [?]
		phone_mod_data[phone_id].line_data[line].locked=true;
		break;

	case phUSstate.usedElsewhere: // Used Elsewhere [small cross]
		if(line==last_used_line.line_id && phone_id==last_used_line.phone_id)
			us_InitiatePhoneFrClose();
		icon=icons.CrossSml;
		break;
	case phUSstate.busyAll: // Busy All [big cross]
		if (line == last_used_line.line_id && phone_id == last_used_line.phone_id)
			us_InitiatePhoneFrClose();
		icon=icons.CrossBig;
		// we can't depend on phSendDropLine being called in this case, so do it ourselves
		markLineAsFree(phone_id, line);
		break;
	case phUSstate.sharedUnavailable: // Shared Unavailable [small cross?]
		// seems like this state is always sent for one of our hybrids
		// when it's being used on another; hence it's ignorable
		icon=-1; // bypass displaying it
		break;

	default:
		if(line==last_used_line.line_id && phone_id==last_used_line.phone_id)
			us_InitiatePhoneFrClose();
		break;
	}

	var led;
	if(phone_mod_data[phone_id].mode24)
	{
	// display the line number instead of an empty icon
		if(icon==0xFF) icon=0;
		led=line;
	}
	else
		led=line+12;

	if(icon>=0)
	{
	// 1 means "immediate sync mode"---i.e., skip preceding sub-icons, go
	// straight to the animation part; 0 means "display the icon with all
	// the sub-icons in the normal way, not skipping anything"...
		mod.setPhoneIcon(phone_id, led, icon, elapsedTimeSec>0?1:0);

	// we need to remember the status for the line, at least, whether it's ringing in.
		phone_mod_data[phone_id].line_data[line].state=lineState;
	}

// for fixed hybrids we additionally try to make them look as close to Europhone channels as we can...
	var fixed_hybrid=phone_mod_data[phone_id].line_data[line].fixed_on_hyb;
	if(fixed_hybrid>=0)
	{
		var faders=phone_mod_data[phone_id].hybrid_data[fixed_hybrid].assignedfaders;

	// replace all directional icons with just a plain down-arrow
		switch(icon)
		{
		case icons.ArrowLt:
		case icons.ArrowRt:
		case icons.LockLt:
		case icons.LockRt:
			icon=icons.ArrowDn;
			break;
		}

		for(var i in faders)
			mod.setChnIcon(faders[i], icon);
	}
}

function phSendSeizeLine(phone_id, line, hybrid)
{
	if(!(phone_id in phUS_ConnInfo))
	{
		if(phone_id in nx_data)
		{
			var nx=nx_data[phone_id];
		//	log.info_f("DBG: NXPhone: line %d siezed on hybrid %d", line, hybrid);
		}
		return;
	}

	var phUS=phUS_ConnInfo[phone_id];

	if(phUS.state!='working') return;

	last_used_line.line_id =line;
	last_used_line.phone_id=phone_id;

	markLineAsUsed(phone_id, line, hybrid);
	us_DialoutFrOpen();
}

function phSendRingLine(phone_id, callerId, line)
{
	if(!(phone_id in phUS_ConnInfo))
	{
		if(phone_id in nx_data)
		{
			var nx=nx_data[phone_id];
			var n, h, cnx;

			if(nx.state!='working') return;
			if(!(line in nx.lines)) return;	// not our line

			n=nx.lines[line];	// hybrid number
			if(!(n in nx.hybrids)) return;		// not supposed to happen
			h=nx.hybrids[n];
			cnx=ch_data[h.fader].nx_phone;
			if(cnx==null) return;

		//	log.info_f("DBG: NXPhone: incoming call on line %d", line);
			showRingingFr(h.fader); // show ringing indication on screen
 			if(ringing_fr_channel==h.fader)
 				gui_indi.updateEUring(null, "Ringing in from "+callerId);
		}
		return;
	}

	var phUS=phUS_ConnInfo[phone_id];

	if(phUS.state!='working') return;
//	log.info_f("Phone [%2d]: SendRingLine('%s', %d)", phone_id, callerId, line);
}

function phSendDropLine(phone_id, callerId, line)
{
	if(!(phone_id in phUS_ConnInfo))
	{
		if(phone_id in nx_data)
		{
			var nx=nx_data[phone_id];
		//	log.info_f("DBG: NXPhone: call dropped line %d", line);
		}
		return;
	}

	var phUS=phUS_ConnInfo[phone_id];

	if(phUS.state!='working') return;
//	log.info_f("Phone [%2d]: SendDropLine('%s', %d)", phone_id, callerId, line);
	markLineAsFree(phone_id, line);
}

function phSendAnswerLine(phone_id, callerId, line, hybrid)
{
	if(!(phone_id in phUS_ConnInfo))
	{
		if(phone_id in nx_data)
		{
			var nx=nx_data[phone_id];
		//	log.info_f("DBG: NXPhone: call on line %d answered on hybrid %d", line, hybrid);
		}
		return;
	}

	var phUS=phUS_ConnInfo[phone_id];

	if(phUS.state!='working') return;
//	log.info_f("Phone [%2d]: SendAnswerLine('%s', %d, %d)", phone_id, callerId, line, hybrid);
}

function phSendSelectLine(phone_id, callerId, line, hybrid)
{
	if(!(phone_id in phUS_ConnInfo))
	{
		if(phone_id in nx_data)
		{
			var nx=nx_data[phone_id];
		//	log.info_f("DBG: NXPhone: call on line %d selected on hybrid %d", line, hybrid);
		}
		return;
	}

	var phUS=phUS_ConnInfo[phone_id];

	if(phUS.state!='working') return;
//	log.info_f("Phone [%2d]: SendSelectLine(%s, %d, %d)", phone_id, callerId, line, hybrid);

	last_used_line.line_id = line;
	last_used_line.phone_id = phone_id;

	markLineAsUsed(phone_id, line, hybrid);
	us_LastcallFrOpen(callerId);
}

function phSendDialKey(phone_id, sym)
{
	if(!(phone_id in phUS_ConnInfo))
	{
		if(phone_id in nx_data)
		{
			var nx=nx_data[phone_id];

			if(dialed_nr.length<25) dialed_nr+=String.fromCharCode(sym);
			gui_indi.updateEUdial(null, null, dialed_nr, null);
		}
		return;
	}

	var phUS=phUS_ConnInfo[phone_id];
	if(phUS.state!='working') return;
//	log.info_f("Phone [%2d]: SendDialKey(%d)", phone_id, sym);

// check range
	if((sym<48 || sym>57) && sym!=35 && sym!=42) return; // not a decimal digit, # or *
	if(dialed_nr.length<25)
		dialed_nr+=String.fromCharCode(sym);
	gui_indi.updateUSdial(dialed_nr);
}

function phSendText(phone_id, textType, text, line)
{
	if(!(phone_id in phUS_ConnInfo))
	{
		if(phone_id in nx_data)
		{
			var nx=nx_data[phone_id];
		//	log.info_f("DBG: NXPhone: send text '%s' of type %d for line %d", text, textType, line);
		}
		return;
	}

	var phUS=phUS_ConnInfo[phone_id];
	if(phUS.state!='working') return;
//	log.info_f("Phone [%2d]: SendText(%d, %s, %d)", phone_id, textType, text, line);
}

// ------------------------- * Divider * -------------------------

function us_DialoutFrOpen()
{
	dialed_nr="";
	gui_indi.updateUSdial(dialed_nr);
}

// closes the dialout frame if it is up
function us_DialoutFrClose()
{
	gui_indi.closeUSdial();
	dialed_nr="";
}

function us_LastcallFrOpen(callerId)
{
	gui_indi.openUScall(callerId);
}

// closes the lastcall frame if it is up
function us_LastcallFrClose()
{
	gui_indi.closeUScall();
}

// closes the dialout and lastcall frame
function us_InitiatePhoneFrClose()
{
	us_DialoutFrClose();
	us_LastcallFrClose();
}

// resets the "Take Line" message associated with some fader
function resetTakeLine()
{
//	log.info_f("resetTakeLine: phone_id = %d", this.pi);

	var takelinehybrid = phone_mod_data[this.pi].takelinehybrid;

	phone_mod_data[this.pi].takelineTO = null;
	phone_mod_data[this.pi].takelinehybrid = -1;

	if (takelinehybrid >= 0)
		showText(this.pi, takelinehybrid);
}

// changes which faders display the "Take Line" text and which hybrid is ready for phone line assignment
function changeTakeLine(phone_id, hybrid)
{
//	log.info_f("changeTakeLine: phone_id = %d, hybrid = %d", phone_id, hybrid);

	if (phone_mod_data[phone_id].num_mon_faders > 0) // if fader monitoring is in effect, don't allow Take Line-s
		return;

	phone_mod_data[phone_id].takelineTO = host.resetTimeout(phone_mod_data[phone_id].takelineTO, takelineTOval, resetTakeLine, { pi : phone_id });

	var takelinehybrid = phone_mod_data[phone_id].takelinehybrid;
	phone_mod_data[phone_id].takelinehybrid = hybrid;
	showText(phone_id, hybrid);

	if (takelinehybrid >= 0 && takelinehybrid != hybrid)
		showText(phone_id, takelinehybrid);
}

// marks the line as used on a particular hybrid, changes what the faders for the hybrid should display
function markLineAsUsed(phone_id, line, hybrid)
{
//	log.info_f("markLineAsUsed: line = %d, hybrid = %d", line, hybrid);

	if (hybrid != 0)
		phone_mod_data[phone_id].last_used_seq_hybrid = hybrid;

	phone_mod_data[phone_id].line_data[line].used_on_hyb = hybrid;

// if we assigned a line to what used to be the "Take Line Hybrid", then reset it
	if (hybrid == phone_mod_data[phone_id].takelinehybrid)
	{
		phone_mod_data[phone_id].takelineTO=host.clearTimeout(phone_mod_data[phone_id].takelineTO);
		phone_mod_data[phone_id].takelinehybrid = -1;
	}

	showText(phone_id, hybrid);
}

function markLineAsFree(phone_id, line)
{
	var hybrid = phone_mod_data[phone_id].line_data[line].used_on_hyb;

	phone_mod_data[phone_id].line_data[line].used_on_hyb = -1;
	showText(phone_id, hybrid);
}

// tests whether attaching/using a line on a hybrid is acceptable w.r.t. lines already locked and the "mashing" attribute
function mayAttach(phone_id, line, hybrid)
{
	var retval = false;

	if (getFirstFader(phone_id, hybrid) != null                            // the hybrid ain't faderless, and
	    && (phone_mod_data[phone_id].line_data[line].used_on_hyb == hybrid // the line is already used on it
		|| getFirstLockedLine(phone_id, hybrid) < 0                    // or no lines are locked on it
		|| phone_mod_data[phone_id].hybrid_data[hybrid].mashing))      // or mashing is allowed
		retval = true;	                                               // then it's OK to attach/use it

	//	log.info_f("mayAttach: line = %d, hybrid = %d, result = %s", line, hybrid, retval ? "true" : "false");
	return retval;
}

// posts the line button to the phone box, thus associating a line with a hybrid, if it is OK to do so
function tryPostLineButton(phone_id, line, hybrid)
{
	//	log.info_f("tryPostLineButton: line = %d, hybrid = %d", line, hybrid);

	if (mayAttach(phone_id, line, hybrid))
		phoneUS.PostLineButton(phone_id, line, hybrid);
}

// finds the best hybrid among the sequenceable hybrids ("Hybrid2", "Hybrid3", etc.), except the fixed-line ones
function findBestHybrid(phone_id, line)
{
	var phUS=phone_mod_data[phone_id];
	var retval=-1; // assume no suitable hybrid will be found
	var hydrid;

// we go from hybrid 2 till the last hybrid
	for(hydrid=1; hydrid<phUS.numhybrids; hydrid++)
	{
	// skip fixed-line hybrids, they have special rules
		if(phUS.hybrid_data[hydrid].fixed_on_line>=0) continue;

		if(mayAttach(phone_id, line, hydrid))
		{
			retval=hydrid;
			break;
		}
	}

//	log.info_f("findBestHybrid: line = %d, result = %d", line, retval);
	return retval;
}

// Shows the source name of the fader on the text display unless there are any lines
// assigned to it, in which case it shows the lowest-numbered line. However, if a "take-line"
// condition is in effect, it displays "Take Line" and nothing else.
function showText(phone_id, hybrid)
{
	//	log.info_f("showText: hybrid = %d", hybrid);

	if (hybrid < 0) // this happens sometimes, e.g., as a result of markLineAsFree() when line ain't used
		return;

	var text = "";

	if (phone_mod_data[phone_id].takelinehybrid == hybrid)
		text = "Take Line";
	else
	{
		var line = getFirstUsedLine(phone_id, hybrid);

		if (line >= 0) // found a line that's used here
			text = host.sprintf("Line %d", line + 1);
	}

	var faders = phone_mod_data[phone_id].hybrid_data[hybrid].assignedfaders;
	var fa_chn;

	for (var i in faders)
	{
		fa_chn=faders[i];
		if (text != "")
			mod.setChnText(fa_chn, text, mod.CENTER);
		else
			setChannelDisp(fa_chn);
	}
}

// finds the first line among those used on a hybrid
function getFirstUsedLine(phone_id, hybrid)
{
	var phUS=phone_mod_data[phone_id];
	var line;

	for(line=0; line<phUS.numlines; line++)
		if(phUS.line_data[line].used_on_hyb==hybrid)
			return line;

	return -1;
}

// finds the first line among those that are locked on a hybrid
function getFirstLockedLine(phone_id, hybrid)
{
	var phUS=phone_mod_data[phone_id];
	var line;

	for(line=0; line<phUS.numlines; line++)
		if(phUS.line_data[line].used_on_hyb==hybrid && phUS.line_data[line].locked)
			return line;

	return -1;
}

// finds the first fader (if any) among those assigned to a hybrid
function getFirstFader(phone_id, hybrid)
{
	// since a key press or something that invokes this function can come any time,
	// we must be prepared to deal with it in the absence of initialized phone structs...
	if (phone_mod_data[phone_id] == null)
		return null;

	var faders = phone_mod_data[phone_id].hybrid_data[hybrid].assignedfaders;

	if (faders != null && faders.length > 0)
		return faders[0];

	return null;
}

// ------------------------- * Telos2 Phone System * -------------------------
function ph2Connected(channel) {}
function ph2Disconnected(channel)
{
// set normal icon
	updateChnIcon(channel);
	mod.setChnKeyLed(channel, keys.ChnSet, 0);
}

var ringing_fr_channel=0; // which channel (fader) is the ringing_fr displayed for (0 if none)

function showRingingFr(channel)
{
	dialingTO=host.clearTimeout(dialingTO);
	gui_indi.updateEUring(ch_data[channel].src_name, null);
	ringing_fr_channel=channel;
	dial_fr=false;
	ringingTO=host.resetTimeout(ringingTO, 10000, ringingFrClose);
}

function ringingFrClose()
{
	dialingTO=host.clearTimeout(dialingTO);
	ringing_fr_channel=0;
	dial_fr=false;
	ringingTO=null;
	gui_indi.closeEU();
}

function linestateFilter()
{
	var phEU=ch_data[this.ch].eu_phone;

	if(phEU==null) return;

	phEU.use_linefilter=false;
	ph2SendLineState(this.ch, this.ls)
	phEU.use_linefilter=true;
	phEU.linestateTO=null;
}

/*
** ph2SendLineState
*/
function ph2SendLineState(channel, lineState)
{
	var chn=ch_data[channel];
	var phEU=chn.eu_phone;
	var icon, set_style, drop;

// filter to filter out short changes (READY -> NO LINE)
	if(phEU.use_linefilter)
	{
	// any message cancels previous timeout
		phEU.linestateTO=host.clearTimeout(phEU.linestateTO);
		if(phEU.line_state==phEUstate.idle && (lineState==phEUstate.unknown || lineState==phEUstate.notready))
		{
			phEU.linestateTO=host.setTimeout(5000, linestateFilter, {ch:channel, ls:lineState});
			return;
		}
	}

// store current line state
	phEU.line_state=lineState;

	icon=0; set_style=0; drop=false;
	switch(lineState)
	{
	case phEUstate.unknown:
	case phEUstate.notready:
	default:
		icon=0; //?
		phEU.on_hold=false;
		break;
	case phEUstate.idle:
		icon=icons.Dot;
		drop=true;
		break;
	case phEUstate.inring:
	case phEUstate.accept:
		icon=icons.Ringing;
		phEU.on_hold=false;
		break;
	case phEUstate.connect: // On Air [Arrow down]
		icon=icons.ArrowDn;
		set_style=255;
		phEU.on_hold=false;
		if(dial_fr)
		{
			gui_indi.updateEUdial(null, "Connected", null, "");
		// close when connected or disconnected
			dialingTO=host.resetTimeout(dialingTO, 3000, dialingFrClose);
		}
		break;
	case phEUstate.disc:
		if(dial_fr)
		{
			gui_indi.updateEUdial(null, "Disconnected", null, "");
		// close when connected or disconnected
			dialingTO=host.resetTimeout(dialingTO, 3000, dialingFrClose);
		}
		break;
	case phEUstate.seizepend:	// seize
		icon=icons.RectSpin2;
		if(dial_fr)
		{
			gui_indi.updateEUdial(null, "Dialing...", null, "");
		// when starts dialing, then don't close dialout frame automatically anymore
			dialingTO=host.clearTimeout(dialingTO);
		}
		break;
	case phEUstate.proceed:
		icon=icons.RectSpin2;
		if(dial_fr) gui_indi.updateEUdial(null, "Proceeding...", null, "");
		break;
	case phEUstate.outring:
		icon=icons.RectSpin2;
		if(dial_fr) gui_indi.updateEUdial(null, "Ringing...", null, "");
		break;
	case phEUstate.suspend: // Hold [box in frame animation]
		icon=icons.OnHold;
		phEU.on_hold=true;
		drop=true;
		break;
	case phEUstate.resume: // Resume hold (turn off hold button light)
		phEU.on_hold=false;
		break;
	}

	if(lineState==phEUstate.inring)
		showRingingFr(channel); // show ringing indication on screen
	else if(channel==ringing_fr_channel)
	{
	// the channel for which we last displayed ringing_fr ain't ringin' no more
		ringingTO=host.clearTimeout(ringingTO); // there must be a timeout related to it
		ringingFrClose();
	}

	if(dial_fr)
	{
		if(lineState==phEUstate.seizepend || lineState==phEUstate.proceed || lineState==phEUstate.outring || // connecting - animated rect
				lineState==phEUstate.connect) // On Air while while dialout frame open - arrow
			mod.setChnIcon(last_used_phone_ch, icon);
		else
			mod.setChnIcon(last_used_phone_ch, icons.RectStat); // typing number - static rect
	}
	else
		mod.setChnIcon(channel, phEU.on_hold?icons.OnHold:icon);
	mod.setChnKeyLed(channel, keys.ChnSet, set_style);
	if(drop) chnLineDroped(channel);
}

function ph2SendRingLine(channel, callerId)
{
	if(ringing_fr_channel==channel)
		gui_indi.updateEUring(null, "Ringing in from "+callerId);
}

function onNXphoneKeyDown(channel, key)
{
	var chn=ch_data[channel];
	var phNX=chn.nx_phone;

	switch(key)
	{
	case keys.ChnHold:
	// if ringing, then answer first
		if(phNX.line_state==phUSstate.ringing)
		{
			phoneUS.PostLineButton(phNX.chn, phNX.line, phNX.hybrid);
			host.sleep(100); // XXX: not 100% clean
		}
		phoneUS.PostHoldButton(phNX.chn, phNX.hybrid);
		break;
	case keys.ChnSet:
	// Set button drops channel
		phoneUS.PostDropButton(phNX.chn, phNX.hybrid);
		break;

	case keys.ChnON:
	case keys.ChnPreview: // Sent only on Pview OFF->ON transition
		if(phNX.line_state!=phUSstate.onAir) // if we are onAir, we will lock, it's not desired
			phoneUS.PostLineButton(phNX.chn, phNX.line, phNX.hybrid);
		break;
	case keys.ChnOFF:
	// OFF button drops line only if Channel is OFF and not in preview
		if(chn.onsw || chn.preview || chn.armed) break;
		phoneUS.PostDropButton(phNX.chn, phNX.hybrid);
		break;
	}
}

function onT2phoneKeyDown(channel, key)
{
	var chn=ch_data[channel];
	var phEU=chn.eu_phone;

	switch(key)
	{
	case keys.ChnHold:
	// if ringing, then answer first
		if(phEU.line_state==phEUstate.inring)
		{
			phoneEU.PostLineButton(channel);
			host.sleep(100); // XXX: not 100% clean
		}
		phoneEU.PostHoldButton(channel);
		break;
	case keys.ChnSet:
	// Set button drops channel
		phoneEU.PostDropButton(channel);
		break;
	case keys.ChnON:
	case keys.ChnPreview: // Sent only on Pview OFF->ON transition
		if(phEU.line_state==phEUstate.inring) phoneEU.PostLineButton(channel);
		if(phEU.on_hold) phoneEU.PostResumeButton(channel);
		break;
	case keys.ChnOFF:
	// OFF button drops line only if Channel is OFF and not in preview
		if(chn.onsw || chn.preview || chn.armed) break;
		phoneEU.PostDropButton(channel);
		break;
	}
}

/*
** onEUphoneKeyDown
*/
function onEUphoneKeyDown(channel, key)
{
	var chn=ch_data[channel];

	if(chn.nx_phone!=null)
		onNXphoneKeyDown(channel, key);
	else if(chn.eu_phone!=null)
		onT2phoneKeyDown(channel, key);
}

function onTelos2DialKeyDown(key)
{
	if(last_used_phone_ch==null) return;

	var chn=ch_data[last_used_phone_ch];
	var sym=phDTMFtoSym[key];
	var nx=chn.nx_phone;

// show dialout frame
	if(!dial_fr)
	{
		ringingTO=host.clearTimeout(ringingTO);
		ringing_fr_channel=0;
		dialed_nr="";
		dial_fr=true;
		if(nx==null) mod.setChnIcon(last_used_phone_ch, icons.RectStat);
		gui_indi.updateEUdial(chn.src_name, dialed_nr, "READY", nx==null?"Press ENTER or # to dial":"");
	}

	if(nx!=null)
	{
	// Nx12 or compatible
		phoneUS.PostDialKey(nx.chn, sym);
	}
	else
	{
		if(sym==35) // ASCII of '#'
			sendDialedNumber();
		else if(dialed_nr.length<25)
		{
			dialed_nr+=String.fromCharCode(sym);
			gui_indi.updateEUdial(null, dialed_nr, null, null);
		}
	}

	gui_indi.animEUdial(key==0?10:(key==11?key:key-1));
// update timeout for closing
	updateDialCloseTO();
}

function onTelos2DialKeyUp(key)
{
	gui_indi.animEUdial(-1);
}

function sendDialedNumber()
{
	if(dial_fr && last_used_phone_ch!=null && ch_data[last_used_phone_ch].nx_phone==null && dialed_nr!="")
		phoneEU.CallNumber(last_used_phone_ch, dialed_nr);
}

function remLastKey()
{
	if(dial_fr)
	{
		if(dialed_nr.length>0)
			dialed_nr=dialed_nr.substr(0, dialed_nr.length-1);
	// FIXME: will not work with Nx12
		gui_indi.updateEUdial(null, dialed_nr, null, null);
	// update timeout for closing
		updateDialCloseTO();
	}
}

function updateDialCloseTO()
{
// update timeout for closing
	dialingTO=host.resetTimeout(dialingTO, 10000, dialingFrClose);
}

function dialingFrClose()
{
	ringingTO=host.clearTimeout(ringingTO);
	ringing_fr_channel=0;
	dial_fr=false;
	dialingTO=null;
	gui_indi.closeEU();
	if(last_used_phone_ch!=null && ch_data[last_used_phone_ch].eu_phone!=null)
		ph2SendLineState(last_used_phone_ch, ch_data[last_used_phone_ch].eu_phone.line_state);
}

// ------------------------- * Telos VX protocol logic * -------------------------

function vxLineState(line_id, state, hybrid)
{
	var icon=255;
	if(line_id<0 || line_id>=12) return;
	switch(state)
	{
	case 'NONE':
	default:
		break;
	case 'IDLE':
	case 'SEIZED':
		icon=icons.Dot;
		break;
	case 'RINGING_IN':
		icon=icons.Ringing;
		break;
	case 'ON_AIR':
		icon=(hybrid==1)?icons.ArrowLt:icons.ArrowRt;
		break;
	case 'ON_AIR_LOCKED':
		icon=(hybrid==1)?icons.LockLt:icons.LockRt;
		break;
	case 'ON_HANDSET':
		icon=icons.Handset;
		break;
	case 'ON_HOLD':
		icon=icons.OnHold;
		break;
	case 'ON_HOLD_READY':
		icon=icons.CheckBig;
		break;
	case 'BUSY':
		icon=icons.CrossBig;
		break;
	case 'SEIZED_EW':
	case 'ON_HANDSET_EW':
	case 'USED_EW':
		icon=icons.CrossSml;
		break;
	}
	mod.setPhoneIcon(1, line_id+12, icon, 0);
}


// ------------------------- * Nx12 or compatible Euro mode * -------------------------
const nx_numchn=4;
// this variable stores nx conection properties
var nx_data=[];

/*
** phNxRegisterFader
**
** returns US phone connection number, or -1 if failed
*/
function phNxRegisterFader(chn, host, hybrid, line, sh_skip)
{
	var n, nx, idle;

/* search for the phone channel to reuse */
	nx=null; idle=null;
	for(n=0; n<nx_numchn; n++)
	{
		if(!((16+n) in nx_data) || nx_data[16+n].state=='idle')
		{
			if(idle==null)
			{
				nx_data[16+n]={chn: 16+n, state: 'idle'};
				idle=nx_data[16+n];
			}
			continue;	// idle connection
		}

		nx=nx_data[16+n];
		if(nx.host==host && nx.sh_skip==sh_skip) // host match
			break; /* found */
		nx=null;
	}
	if(nx==null) nx=idle;
	if(nx==null)
	{
		log.warning_f("phone: no free Nx connection for %s/%d:%d in ch%02d", host, hybrid, line, chn);
		return -1; // no free connection
	}

	if(nx.state=='idle')
	{
	/* grab nx connection */
		nx.host=host;
		nx.shname=null;
		nx.hostName=null;
		nx.hybrids=[];
		nx.lines=[];
		nx.sh_skip=sh_skip;
		nx.sh_skipped=0;
		nx.state='conhost';

		phoneUS.OpenConnection(nx.chn, nx.host, "telos", "telos");
	}

/* check for the same host */
	if(nx.host!=host || nx.sh_skip!=sh_skip) return -1;

/* register hybrid */
	if(hybrid in nx.hybrids)
	{
		log.warning_f("phone: Nx hybrid %d @ %s is in use on ch%02d", hybrid, host, nx.hybrids[hybrid].fader);
		return -1;
	}
/* check if line is free */
	if(line in nx.lines)
	{
		log.warning_f("phone: Nx line %d @ %s is in use on ch%02d", line, host, nx.hybrids[nx.lines[line]].fader);
		return -1;
	}
	nx.lines[line]=hybrid;
	nx.hybrids[hybrid]=
	{
		fader: chn,		// fader the hybrid is bound to
		line:  line		// phone line bound to hybrid
	};
// update information about our line if we are joining existing connection
	if(nx.state=='working') phoneUS.PostUpdateAll(nx.chn);

	return nx.chn;
}

/*
** phNxUnregisterFader
*/
function phNxUnregisterFader(id, chn, hybrid, line)
{
	var ind;
	var nx;

	if(!id in nx_data) return;
	nx=nx_data[id];

	if(!(hybrid in nx.hybrids)) return;
	if(nx.hybrids[hybrid].fader!=chn || nx.hybrids[hybrid].line!=line) return;

/* remove hybrid and line */
	delete nx.hybrids[hybrid];
	delete nx.lines[line];

/* close connection if no more hybrids left */
	for(ind in nx.hybrids)
	{
		if('fader' in nx.hybrids[ind]) return;
	}

/* no more hybrids left */
	phoneUS.CloseConnection(nx.chn);
	delete nx_data[id];
}

function phoneMuteRinger(mute)
{
	var ind;
	var nx, us;

	if(phUS_muteRinger==mute) return;
	phUS_muteRinger=mute;

	for(ind in nx_data)
	{
		nx=nx_data[ind];
		if(nx.state=='working') phoneUS.MuteRinger(nx.chn, phUS_muteRinger);
	}
	for(ind in phUS_ConnInfo)
	{
		us=phUS_ConnInfo[ind];
		if(us.state=='working') phoneUS.MuteRinger(ind, phUS_muteRinger);
	}
}
